from langchain_openai.chat_models import ChatOpenAI

from ragas.llms.base import BaseRagasLLM, LangchainLLMWrapper

__all__ = [
    "BaseRagasLLM",
    "LangchainLLMWrapper",
    "llm_factory",
]


def llm_factory(model="gpt-3.5-turbo-16k") -> BaseRagasLLM:
    return LangchainLLMWrapper(ChatOpenAI(model=model))



from __future__ import annotations

import logging
import typing as t
from abc import abstractmethod
from dataclasses import dataclass, field

from fsspec.exceptions import asyncio
from langchain_core.pydantic_v1 import BaseModel
from numpy.random import default_rng

from ragas.llms import BaseRagasLLM
from ragas.llms.prompt import Prompt
from ragas.testset.docstore import Direction, DocumentStore, Node
from ragas.testset.filters import EvolutionFilter, NodeFilter, QuestionFilter
from ragas.testset.prompts import (
    compress_question_prompt,
    multi_context_question_prompt,
    question_answer_prompt,
    reasoning_question_prompt,
    seed_question_prompt,
)

rng = default_rng()
logger = logging.getLogger(__name__)


@dataclass
class CurrentNodes:
    root_node: Node
    nodes: t.List[Node] = field(default_factory=list)


# (question, current_nodes, evolution_type)
EvolutionOutput = t.Tuple[str, CurrentNodes, str]


class DataRow(BaseModel):
    question: str
    context: str
    answer: str
    evolution_type: str


@dataclass
class Evolution:
    generator_llm: BaseRagasLLM = t.cast(BaseRagasLLM, None)
    docstore: t.Optional[DocumentStore] = None
    node_filter: t.Optional[NodeFilter] = None
    question_filter: t.Optional[QuestionFilter] = None
    max_tries: int = 5

    @staticmethod
    def merge_nodes(nodes: CurrentNodes) -> Node:
        return Node(
            doc_id="merged", page_content="\n".join(n.page_content for n in nodes.nodes)
        )

    async def aretry_evolve(
        self, current_tries: int, current_nodes: CurrentNodes, update_count: bool = True
    ) -> EvolutionOutput:
        if update_count:
            current_tries += 1
        logger.info("retrying evolution: %s times", current_tries)
        if current_tries > self.max_tries:
            # TODO: make this into a custom exception
            raise ValueError("Max tries reached")
        return await self._aevolve(current_tries, current_nodes)

    def _transform_question(self, prompt: Prompt, question: str) -> str:
        assert self.generator_llm is not None, "generator_llm cannot be None"

        results = self.generator_llm.generate_text(
            prompt=prompt.format(question=question)
        )
        return results.generations[0][0].text.strip()

    def _get_more_adjacent_nodes(self, current_nodes: CurrentNodes):
        """
        if the evolutions doesn't have enough nodes to frame a question, get more nodes
        """
        assert self.docstore is not None, "docstore cannot be None"

        # get more nodes from above the context window
        prev_adjacent_node = self.docstore.get_adjacent(
            current_nodes.nodes[0], Direction.PREV
        )
        if prev_adjacent_node is None:
            # get more nodes from below the context window
            next_adjacent_node = self.docstore.get_adjacent(
                current_nodes.nodes[-1], Direction.NEXT
            )
            if next_adjacent_node is not None:
                # add next nodes towards the end
                current_nodes.nodes.append(next_adjacent_node)
            else:
                # retry with new base node
                nodes = self.docstore.get_random_nodes(k=1)
                return CurrentNodes(root_node=nodes[0], nodes=nodes)
        else:
            # add prev nodes in index 0
            current_nodes.nodes.insert(0, prev_adjacent_node)

        return current_nodes

    def evolve(self, current_nodes: CurrentNodes) -> DataRow:
        return asyncio.get_event_loop().run_until_complete(self.aevolve(current_nodes))

    async def aevolve(self, current_nodes: CurrentNodes) -> DataRow:
        # init tries with 0 when first called
        current_tries = 0
        (
            evolved_question,
            current_nodes,
            evolution_type,
        ) = await self._aevolve(current_tries, current_nodes)

        return self.generate_datarow(
            question=evolved_question,
            current_nodes=current_nodes,
            evolution_type=evolution_type,
        )

    @abstractmethod
    async def _aevolve(
        self, current_tries: int, current_nodes: CurrentNodes
    ) -> EvolutionOutput:
        ...

    def generate_datarow(
        self,
        question: str,
        current_nodes: CurrentNodes,
        evolution_type: str,
    ):
        assert self.generator_llm is not None, "generator_llm cannot be None"

        merged_nodes = self.merge_nodes(current_nodes)
        results = self.generator_llm.generate_text(
            prompt=question_answer_prompt.format(
                question=question, context=merged_nodes.page_content
            )
        )
        answer = results.generations[0][0].text.strip()
        logger.debug("answer generated: %s", answer)

        if answer == "-1":
            answer = None

        return DataRow(
            question=question,
            context=merged_nodes.page_content,
            answer="" if answer is None else answer,
            evolution_type=evolution_type,
        )


@dataclass
class SimpleEvolution(Evolution):
    async def _aevolve(
        self, current_tries: int, current_nodes: CurrentNodes
    ) -> EvolutionOutput:
        assert self.docstore is not None, "docstore cannot be None"
        assert self.node_filter is not None, "node filter cannot be None"
        assert self.generator_llm is not None, "generator_llm cannot be None"
        assert self.question_filter is not None, "question_filter cannot be None"

        merged_node = self.merge_nodes(current_nodes)
        passed = await self.node_filter.afilter(current_nodes.root_node)
        if not passed["score"]:
            nodes = self.docstore.get_random_nodes(k=1)
            new_current_nodes = CurrentNodes(root_node=nodes[0], nodes=nodes)
            return await self.aretry_evolve(
                current_tries, new_current_nodes, update_count=False
            )

        results = self.generator_llm.generate_text(
            prompt=seed_question_prompt.format(context=merged_node.page_content)
        )
        seed_question = results.generations[0][0].text
        # NOTE: might need improvement
        # select only one seed question here
        is_valid_question = await self.question_filter.afilter(seed_question)
        if not is_valid_question:
            # get more context to rewrite question
            current_nodes = self._get_more_adjacent_nodes(current_nodes)
            # retry with new nodes added
            return await self.aretry_evolve(current_tries, current_nodes)
        else:
            # if valid question
            return seed_question, current_nodes, "simple"

    def __hash__(self):
        return hash(self.__class__.__name__)


@dataclass
class ComplexEvolution(Evolution):
    se: t.Optional[SimpleEvolution] = field(default=None, repr=False)
    evolution_filter: t.Optional[EvolutionFilter] = field(default=None, repr=False)

    def init_evolution(self):
        # init simple evolution to get seed question
        self.se = SimpleEvolution(
            generator_llm=self.generator_llm,
            docstore=self.docstore,
            node_filter=self.node_filter,
            question_filter=self.question_filter,
        )
        # init evolution filter with critic llm from another filter
        assert self.node_filter is not None, "node filter cannot be None"
        self.evolution_filter = EvolutionFilter(self.node_filter.llm)


@dataclass
class MultiContextEvolution(ComplexEvolution):
    async def _aevolve(
        self, current_tries: int, current_nodes: CurrentNodes
    ) -> EvolutionOutput:
        assert self.docstore is not None, "docstore cannot be None"
        assert self.generator_llm is not None, "generator_llm cannot be None"
        assert self.question_filter is not None, "question_filter cannot be None"
        assert self.se is not None, "simple evolution cannot be None"

        simple_question, _, _ = await self.se._aevolve(current_tries, current_nodes)
        logger.debug(
            "[MultiContextEvolution] simple question generated: %s", simple_question
        )

        # find a similar node and generate a question based on both
        similar_node = self.docstore.get_similar(current_nodes.root_node)[0]
        prompt = multi_context_question_prompt.format(
            question=simple_question,
            context1=current_nodes.root_node.page_content,
            context2=similar_node,
        )
        results = await self.generator_llm.agenerate_text(prompt=prompt)
        question = results.generations[0][0].text.strip()
        logger.debug(
            "[MultiContextEvolution] multicontext question generated: %s", question
        )

        # compress the question
        compressed_question = self._transform_question(
            prompt=compress_question_prompt, question=question
        )
        logger.debug(
            "[MultiContextEvolution] multicontext question compressed: %s", question
        )

        if not await self.question_filter.afilter(compressed_question):
            # retry
            current_nodes = self.se._get_more_adjacent_nodes(current_nodes)
            return await self.aretry_evolve(current_tries, current_nodes)

        assert self.evolution_filter is not None, "evolution filter cannot be None"
        if not await self.evolution_filter.afilter(
            simple_question, compressed_question
        ):
            # retry
            current_nodes = self.se._get_more_adjacent_nodes(current_nodes)
            return await self.aretry_evolve(current_tries, current_nodes)

        return compressed_question, current_nodes, "multi_context"

    def __hash__(self):
        return hash(self.__class__.__name__)


@dataclass
class ReasoningEvolution(ComplexEvolution):
    async def _aevolve(
        self, current_tries: int, current_nodes: CurrentNodes
    ) -> EvolutionOutput:
        assert self.generator_llm is not None, "generator_llm cannot be None"
        assert self.question_filter is not None, "question_filter cannot be None"
        assert self.se is not None, "simple evolution cannot be None"

        simple_question, _, _ = await self.se._aevolve(current_tries, current_nodes)
        logger.debug(
            "[ReasoningEvolution] simple question generated: %s", simple_question
        )

        result = await self.generator_llm.agenerate_text(
            prompt=reasoning_question_prompt.format(
                question=simple_question, context=current_nodes.root_node.page_content
            )
        )
        reasoning_question = result.generations[0][0].text.strip()
        #
        # compress the question
        compressed_question = self._transform_question(
            prompt=compress_question_prompt, question=reasoning_question
        )
        logger.debug(
            "[ReasoningEvolution] multicontext question compressed: %s",
            reasoning_question,
        )

        if not await self.question_filter.afilter(compressed_question):
            # retry
            current_nodes = self.se._get_more_adjacent_nodes(current_nodes)
            return await self.aretry_evolve(current_tries, current_nodes)

        assert self.evolution_filter is not None, "evolution filter cannot be None"
        if not await self.evolution_filter.afilter(
            simple_question, compressed_question
        ):
            # retry
            current_nodes = self.se._get_more_adjacent_nodes(current_nodes)
            logger.debug(
                "evolution_filter failed, retrying with %s", len(current_nodes.nodes)
            )
            return await self.aretry_evolve(current_tries, current_nodes)

        return reasoning_question, current_nodes, "reasoning"

    def __hash__(self):
        return hash(self.__class__.__name__)


simple = SimpleEvolution()
multi_context = MultiContextEvolution()
reasoning = ReasoningEvolution()



from langchain_core.prompts import HumanMessagePromptTemplate

from ragas.llms.prompt import Prompt

seed_question_prompt = Prompt(
    name="seed_question",
    instruction="""Generate a question from given context satisfying the rules given below:
    2.The question should be framed such that it must be clearly understood without providing context.
    3.The question should be fully answerable from information present in given context.""",
    examples=[
        {
            "context": "The Eiffel Tower in Paris was originally intended as a temporary structure, built for the 1889 World's Fair. It was almost dismantled in 1909 but was saved because it was repurposed as a giant radio antenna.",
            "output": "Who built the Eiffel Tower?",
        },
    ],
    input_keys=["context"],
    output_key="output",
    output_type="string",
    language="english",
)


reasoning_question_prompt = Prompt(
    name="reasoning_question",
    instruction="""Complicate the given question by rewriting question into a multi-hop reasoning question based on the provided context.
    Answering the question should require the reader to make multiple logical connections or inferences using the information available in given context.
    Rules to follow when rewriting question:
    1. Ensure that the rewritten question can be answered entirely from the information present in the contexts.
    2. Do not frame questions that contains more than 15 words. Use abbreviation wherever possible.
    3. Make sure the question is clear and unambiguous.
    4. phrases like 'based on the provided context','according to the context',etc are not allowed to appear in the question.""",
    examples=[
        {
            "question": "What is the capital of France?",
            "context": "France is a country in Western Europe. It has several cities, including Paris, Lyon, and Marseille. Paris is not only known for its cultural landmarks like the Eiffel Tower and the Louvre Museum but also as the administrative center.",
            "output": "Linking the Eiffel Tower and administrative center, which city stands as both?",
        },
        {
            "question": "What does the append() method do in Python?",
            "context": "In Python, lists are used to store multiple items in a single variable. Lists are one of 4 built-in data types used to store collections of data. The append() method adds a single item to the end of a list.",
            "output": "If a list represents a variable collection, what method extends it by one item?",
        },
    ],
    input_keys=["question", "context"],
    output_key="output",
    output_type="string",
    language="english",
)


multi_context_question_prompt = Prompt(
    name="multi_context_question",
    instruction="""
    The task is to rewrite and complicate the given question in a way that answering it requires information derived from both context1 and context2. 
    Follow the rules given below while rewriting the question.
        1. The rewritten question should not be very long. Use abbreviation wherever possible.
        2. The rewritten question must be reasonable and must be understood and responded by humans.
        3. The rewritten question must be fully answerable from information present in context1 and context2. 
        4. Read and understand both contexts and rewrite the question so that answering requires insight from both context1 and context2.
        5. phrases like 'based on the provided context','according to the context?',etc are not allowed to appear in the question.""",
    examples=[
        {
            "question": "What process turns plants green?",
            "context1": "Chlorophyll is the pigment that gives plants their green color and helps them photosynthesize.",
            "context2": "Photosynthesis in plants typically occurs in the leaves where chloroplasts are concentrated.",
            "output": "In which plant structures does the pigment responsible for their verdancy facilitate energy production?",
        },
        {
            "question": "How do you calculate the area of a rectangle?",
            "context1": "The area of a shape is calculated based on the shape's dimensions. For rectangles, this involves multiplying the length and width.",
            "context2": "Rectangles have four sides with opposite sides being equal in length. They are a type of quadrilateral.",
            "output": "What multiplication involving equal opposites yields a quadrilateral's area?",
        },
    ],
    input_keys=["question", "context1", "context2"],
    output_key="output",
    output_type="string",
    language="english",
)

conditional_question_prompt = Prompt(
    name="conditional_question",
    instruction="""Rewrite the provided question to increase its complexity by introducing a conditional element.
    The goal is to make the question more intricate by incorporating a scenario or condition that affects the context of the question.
    Follow the rules given below while rewriting the question.
        1. The rewritten question should not be longer than 25 words. Use abbreviation wherever possible.
        2. The rewritten question must be reasonable and must be understood and responded by humans.
        3. The rewritten question must be fully answerable from information present context.
        4. phrases like 'provided context','according to the context?',etc are not allowed to appear in the question.""",
    examples=[
        {
            "question": "What is the function of the roots of a plant?",
            "context": "The roots of a plant absorb water and nutrients from the soil, anchor the plant in the ground, and store food.",
            "output": "What dual purpose do plant roots serve concerning soil nutrients and stability?",
        },
        {
            "question": "How do vaccines protect against diseases?",
            "context": "Vaccines protect against diseases by stimulating the body's immune response to produce antibodies, which recognize and combat pathogens.",
            "output": "How do vaccines utilize the body's immune system to defend against pathogens?",
        },
    ],
    input_keys=["question", "context"],
    output_key="output",
    output_type="string",
    language="english",
)


compress_question_prompt = Prompt(
    name="compress_question",
    instruction="""Rewrite the following question to make it more indirect and shorter while retaining the essence of the original question.
    The goal is to create a question that conveys the same meaning but in a less direct manner. The rewritten question should shorter so use abbreviation wherever possible.""",
    examples=[
        {
            "question": "What is the distance between the Earth and the Moon?",
            "output": "How far is the Moon from Earth?",
        },
        {
            "question": "What ingredients are required to bake a chocolate cake?",
            "output": "What's needed for a chocolate cake?",
        },
    ],
    input_keys=["question"],
    output_key="output",
    output_type="string",
    language="english",
)


conversational_question_prompt = Prompt(
    name="conversation_question",
    instruction="""Reformat the provided question into two separate questions as if it were to be part of a conversation. Each question should focus on a specific aspect or subtopic related to the original question.
    Follow the rules given below while rewriting the question.
        1. The rewritten question should not be longer than 25 words. Use abbreviation wherever possible.
        2. The rewritten question must be reasonable and must be understood and responded by humans.
        3. The rewritten question must be fully answerable from information present context.
        4. phrases like 'provided context','according to the context?',etc are not allowed to appear in the question.""",
    examples=[
        {
            "question": "What are the advantages and disadvantages of remote work?",
            "output": {
                "first_question": "What are the benefits of remote work?",
                "second_question": "On the flip side, what challenges are encountered when working remotely?",
            },
        }
    ],
    input_keys=["question"],
    output_key="output",
    output_type="json",
    language="english",
)

context_scoring_prompt = Prompt(
    name="score_context",
    instruction="""Given a context, complete the two following tasks and output answer valid json format
1.Evaluate the provided context and assign a numerical score between 0 and 10 based on the following criteria:
    - Award a high score to context that thoroughly delves into and explains concepts.
    - Assign a lower score to context that contains excessive references, acknowledgments, personal information, or other non-essential elements.""",
    examples=[
        {
            "context": "Albert Einstein (14 March 1879 - 18 April 1955) was a German-born theoretical physicist who is widely held to be one of the greatest and most influential scientists of all time.",
            "output": {"score": 6.0},
        }
    ],
    input_keys=["context"],
    output_key="output",
    output_type="json",
    language="english",
)

question_rewrite_prompt = Prompt(
    name="rewrite_question",
    instruction="""Given a context, transform the given question to be clear and standalone by replacing its coreferences with specific details from the context:""",
    examples=[
        {
            "context": "The Eiffel Tower was constructed using iron and was originally intended as a temporary exhibit for the 1889 World's Fair held in Paris. Despite its initial temporary purpose, the Eiffel Tower quickly became a symbol of Parisian ingenuity and an iconic landmark of the city, attracting millions of visitors each year. The tower's design, created by Gustave Eiffel, was initially met with criticism from some French artists and intellectuals, but it has since been celebrated as a masterpiece of structural engineering and architectural design.",
            "question": "Who created the design for the Tower?",
            "output": "Who created the design for the Eiffel Tower?",
        },
        {
            "context": "'Exploring Zero-Shot Learning in Neural Networks' was published by Smith and Lee in 2021, focusing on the application of zero-shot learning techniques in artificial intelligence.",
            "question": "What datasets were used for the zero-shot evaluations in this study?",
            "output": "What datasets were used for the zero-shot evaluations Exploring Zero-Shot Learning in Neural Networks paper?",
        },
    ],
    input_keys=["context", "question"],
    output_key="output",
    output_type="string",
    language="english",
)


filter_question_prompt = Prompt(
    name="filter_question",
    instruction="""Given a question, classify it based on clarity and specificity""",
    examples=[
        {
            "question": "What is the discovery about space?",
            "output": {
                "reason": "The question is too vague and does not specify which discovery about space it is referring to.",
                "verdit": "No",
            },
        },
        {
            "question": "What caused the Great Depression?",
            "output": {
                "reason": "The question is specific and refers to a well-known historical economic event, making it clear and answerable.",
                "verdict": "Yes",
            },
        },
        {
            "question": "What is the keyword that best describes the paper's focus in natural language understanding tasks?",
            "output": {
                "reason": "The question mentions a 'paper' in it without referring it's name which makes it unclear without it",
                "verdict": "No",
            },
        },
        {
            "question": "Who wrote 'Romeo and Juliet'?",
            "output": {
                "reason": "The question is clear and refers to a specific work by name therefore it is clear",
                "verdict": "Yes",
            },
        },
        {
            "question": "What did the study mention?",
            "output": {
                "reason": "The question is vague and does not specify which study it is referring to",
                "verdict": "No",
            },
        },
        {
            "question": "What is the focus of the REPLUG paper?",
            "output": {
                "reason": "The question refers to a specific work by it's name hence can be understood",
                "verdict": "Yes",
            },
        },
        {
            "question": "What is the purpose of the reward-driven stage in the training process?",
            "output": {
                "reason": "The question lacks specific context regarding the type of training process, making it potentially ambiguous and open to multiple interpretations.",
                "verdict": "No",
            },
        },
    ],
    input_keys=["question"],
    output_key="output",
    output_type="json",
    language="english",
)

evolution_elimination_prompt = Prompt(
    name="evolution_elimination",
    instruction="""Check if the given two questions are equal based on following requirements:
    1. They have same constraints and requirements.
    2. They have same depth and breadth of the inquiry.""",
    examples=[
        {
            "question1": "What are the primary causes of climate change?",
            "question2": "What factors contribute to global warming?",
            "output": {
                "reason": "While both questions deal with environmental issues, 'climate change' encompasses broader changes than 'global warming', leading to different depths of inquiry.",
                "verdict": "Not Equal",
            },
        },
        {
            "question1": "How does photosynthesis work in plants?",
            "question2": "Can you explain the process of photosynthesis in plants?",
            "output": {
                "reason": "Both questions ask for an explanation of the photosynthesis process in plants, sharing the same depth, breadth, and requirements for the answer.",
                "verdict": "Equal",
            },
        },
    ],
    input_keys=["question1", "question2"],
    output_key="output",
    output_type="json",
    language="english",
)

question_answer_prompt = Prompt(
    name="answer_formulate",
    instruction="""Answer the question using the information from the given context. Answer '-1' if answer is not present in the context.""",
    examples=[
        {
            "context": """The novel '1984' by George Orwell is set in a dystopian future where the world is divided into three superstates. The story follows the life of Winston Smith, who lives in Oceania, a superstate constantly at war.""",
            "question": "In which superstate does Winston Smith live in the novel '1984'?",
            "answer": "Winston Smith lives in the superstate of Oceania in the novel '1984'.",
        },
        {
            "context": """The novel "Pride and Prejudice" by Jane Austen revolves around the character Elizabeth Bennet and her family. The story is set in the 19th century in rural England and deals with issues of marriage, morality, and misconceptions.""",
            "question": "What year was 'Pride and Prejudice' published?",
            "answer": "-1",
        },
    ],
    input_keys=["context", "question"],
    output_key="answer",
    output_type="string",
    language="english",
)


## TODO: remove this

SEED_QUESTION = HumanMessagePromptTemplate.from_template(
    """
Generate two questions from given context satisfying the rules given below:
    2.The question should be framed such that it must be clearly understood without providing context.
    3.The question should be fully answerable from information present in given context.
    

{demonstration}


Context:
{context}
Questions:"""  # noqa: E501
)

TABLE_QA = HumanMessagePromptTemplate.from_template(
    """
Frame a question from the given table following the rules given below
    - Do no use phrases like 'provided context','provided table' etc in the question
    
Context:
Table 2: Local Library Statistics

Month	New Memberships	Books Loaned	eBooks Downloaded
January	150	1200	950
February	120	1100	1000
March	200	1400	1100

Framed Question from Table: How many books were loaned in January?

Context:
{context}

Framed Question from Table:"""  # noqa: E501
)


REASONING_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
Complicate the given question by rewriting question into a multi-hop reasoning question based on the provided context. 
Answering the question should require the reader to make multiple logical connections or inferences using the information available in given context. 
Rules to follow when rewriting question:
1. Ensure that the rewritten question can be answered entirely from the information present in the contexts.
2. Do not frame questions that contains more than 15 words. Use abbreviation wherever possible.
3. Make sure the question is clear and unambiguous. 
4. phrases like 'based on the provided context','according to the context',etc are not allowed to appear in the question.

Initial Question:
What is the capital of France?

Given Context:
France is a country in Western Europe. It has several cities, including Paris, Lyon, and Marseille. Paris is not only known for its cultural landmarks like the Eiffel Tower and the Louvre Museum but also as the administrative center.

Complicated Multi-Hop Question:
Linking the Eiffel Tower and administrative center, which city stands as both?

Initial Question:
What does the append() method do in Python?

Given Context:
In Python, lists are used to store multiple items in a single variable. Lists are one of 4 built-in data types used to store collections of data. The append() method adds a single item to the end of a list.

Complicated Multi-Hop Question:
If a list represents a variable collection, what method extends it by one item?

Initial Question: 
{question}
Given Context:
{context}

Complicated Multi-Hop Question
"""  # noqa: E501
)

MULTICONTEXT_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
The task is to rewrite and complicate the given question in a way that answering it requires information derived from both context1 and context2. 
Follow the rules given below while rewriting the question.
    1. The rewritten question should not be very long. Use abbreviation wherever possible.
    2. The rewritten question must be reasonable and must be understood and responded by humans.
    3. The rewritten question must be fully answerable from information present in context1 and context2. 
    4. Read and understand both contexts and rewrite the question so that answering requires insight from both context1 and context2.
    5. phrases like 'based on the provided context','according to the context?',etc are not allowed to appear in the question.

Initial Question:
What process turns plants green?

Context1:
Chlorophyll is the pigment that gives plants their green color and helps them photosynthesize.

Context2:
Photosynthesis in plants typically occurs in the leaves where chloroplasts are concentrated.

Complicated Multi-Hop Question:
In which plant structures does the pigment responsible for their verdancy facilitate energy production?

Initial Question:
How do you calculate the area of a rectangle?

Context1:
The area of a shape is calculated based on the shape's dimensions. For rectangles, this involves multiplying the length and width.

Context2:
Rectangles have four sides with opposite sides being equal in length. They are a type of quadrilateral.

Complicated Multi-Hop Question:
What multiplication involving equal opposites yields a quadrilateral's area?


Initial Question:
{question}
context1:
{context1}
context2:
{context2}
Complicated Multi-Hop Question:
"""  # noqa: E501
)


CONDITIONAL_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
Rewrite the provided question to increase its complexity by introducing a conditional element.
The goal is to make the question more intricate by incorporating a scenario or condition that affects the context of the question.
Follow the rules given below while rewriting the question.
    1. The rewritten question should not be longer than 25 words. Use abbreviation wherever possible.
    2. The rewritten question must be reasonable and must be understood and responded by humans.
    3. The rewritten question must be fully answerable from information present context.
    4. phrases like 'provided context','according to the context?',etc are not allowed to appear in the question.

Initial Question:
What is the function of the roots of a plant?

Context:
The roots of a plant absorb water and nutrients from the soil, anchor the plant in the ground, and store food.

Rewritten Question:
What dual purpose do plant roots serve concerning soil nutrients and stability?

Answer:
Plant roots serve a dual purpose by absorbing water and nutrients from the soil, which is vital for the plant's growth, and providing stability by anchoring the plant in the ground.

Example 2:

Initial Question:
How do vaccines protect against diseases?

Context:
Vaccines protect against diseases by stimulating the body's immune response to produce antibodies, which recognize and combat pathogens.

Rewritten Question:
How do vaccines utilize the body's immune system to defend against pathogens?

Initial Question::
{question}
Context:
{context}
Rewritten Question
"""  # noqa: E501
)


COMPRESS_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
Rewrite the following question to make it more indirect and shorter while retaining the essence of the original question. 
The goal is to create a question that conveys the same meaning but in a less direct manner. The rewritten question should shorter so use abbreviation wherever possible.

Original Question:
What is the distance between the Earth and the Moon?

Rewritten Question:
How far is the Moon from Earth?

Original Question:
What ingredients are required to bake a chocolate cake?

Rewritten Question:
What's needed for a chocolate cake?

Original Question:
{question}
Rewritten Question:
"""  # noqa: E501
)


CONVERSATION_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
Reformat the provided question into two separate questions as if it were to be part of a conversation. Each question should focus on a specific aspect or subtopic related to the original question.
question: What are the advantages and disadvantages of remote work?
Reformatted Questions for Conversation: What are the benefits of remote work?\nOn the flip side, what challenges are encountered when working remotely?
question:{question}

Reformatted Questions for Conversation:
"""  # noqa: E501
)

SCORE_CONTEXT = HumanMessagePromptTemplate.from_template(
    """Given a context, complete the two following tasks and output answer valid json format 
1.Evaluate the provided context and assign a numerical score between 0 and 10 based on the following criteria:
    - Award a high score to context that thoroughly delves into and explains concepts.
    - Assign a lower score to context that contains excessive references, acknowledgments, personal information, or other non-essential elements.
2.Check if context contains tables
Context:
Albert Einstein (/ˈaɪnstaɪn/ EYEN-styne;[4] German: [ˈalbɛɐt ˈʔaɪnʃtaɪn] ⓘ; 14 March 1879 – 18 April 1955) was a German-born theoretical physicist who is widely held to be one of the greatest and most influential scientists of all time.
Output:
{{"score":6.0, "is_table_present":false}}
Context:
{context}
Output:"""  # noqa: E501
)

REWRITE_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
Given a context, transform the given question to be clear and standalone by replacing its coreferences with specific details from the context:

Contexts:
The Eiffel Tower was constructed using iron and was originally intended as a temporary exhibit for the 1889 World's Fair held in Paris.
Despite its initial temporary purpose, the Eiffel Tower quickly became a symbol of Parisian ingenuity and an iconic landmark of the city, attracting millions of visitors each year
The tower's design, created by Gustave Eiffel, was initially met with criticism from some French artists and intellectuals, but it has since been celebrated as a masterpiece of structural engineering and architectural design.
Question:
Who created the design for the Tower?
Rewritten question:
Who created the design for the Eiffel Tower?

Contexts:
'Exploring Zero-Shot Learning in Neural Networks' was published by Smith and Lee in 2021, focusing on the application of zero-shot learning techniques in artificial intelligence. 
Question: 
What datasets were used for the zero-shot evaluations in this study?
Rewritten question:
What datasets were used for the zero-shot evaluations Exploring Zero-Shot Learning in Neural Networks paper?


Question:{question}
Context: {context}
Rewritten question:
"""
)

FILTER_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
Classify given question

question: What is the discovery about space?
{{
    "reason":"The question is too vague and does not specify which discovery about space it is referring to."
    "verdit":"No"
}}

question: What caused the Great Depression?
{{
    "reason":"The question is specific and refers to a well-known historical economic event, making it clear and answerable.",
    "verdict":"Yes"
}}

question: What is the keyword that best describes the paper's focus in natural language understanding tasks?
{{
  "reason": "The question mentions a 'paper' in it without referring it's name which makes it unclear without it",
  "verdict": "No"
}}
question: Who wrote 'Romeo and Juliet'?
{{
  "reason": "The question is clear and refers to a specific work by name therefore it is clear",
  "verdict": "Yes"
}}
question: What did the study mention?
{{
  "reason": "The question is vague and does not specify which study it is referring to",
  "verdict": "No"
}}
question: What is the focus of the REPLUG paper?
{{
    "reason": "The question refers to a specific work by it's name hence can be understood", 
    "verdict": "Yes"
}}

question: What is the purpose of the reward-driven stage in the training process?
{{
"reason": "The question lacks specific context regarding the type of training process, making it potentially ambiguous and open to multiple interpretations.",
"verdict": "No"
}}


question: {question}"""  # noqa: E501
)

EVOLUTION_ELIMINATION = HumanMessagePromptTemplate.from_template(
    """\
Check if the given two questions are equal based on following requirements:
1. They have same constraints and requirements.
2. They have same depth and breadth of the inquiry.

Question 1: What are the primary causes of climate change?
Question 2: What factors contribute to global warming?
{{
  "reason": "While both questions deal with environmental issues, 'climate change' encompasses broader changes than 'global warming', leading to different depths of inquiry.",
  "verdict": "Not Equal"
}}
Question 1: How does photosynthesis work in plants?
Question 2: Can you explain the process of photosynthesis in plants?
{{
  "reason": "Both questions ask for an explanation of the photosynthesis process in plants, sharing the same depth, breadth, and requirements for the answer.",
  "verdict": "Equal"
}}
Question 1: {question1}
Question 2: {question2}"""  # noqa: E501
)

ANSWER_FORMULATE = HumanMessagePromptTemplate.from_template(
    """\
Answer the question using the information from the given context. 
question:{question}
context:{context}
answer:
"""  # noqa: E501
)


INFORMAL_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
Rewrite the following question into a casual, conversational form as if it's being asked by someone in an informal setting. 
Keep the core information request intact, without including any additional details or questions.
Formal Question: What are the criteria for Objectives and Key Results?
Casual Rewrite: What should I be looking at when I'm setting up OKRs?
Formal Question: Could you delineate the primary responsibilities of a project manager?
Casual Rewrite: What's the main job of a project manager, in simple terms?
Formal Question: What mechanisms underlie the process of cellular respiration?
Casual Rewrite: How does cellular respiration actually work?
Formal Question:{question}
Casual Rewrite:"""
)

CONTEXT_FORMULATE = HumanMessagePromptTemplate.from_template(
    """Please extract relevant sentences from the provided context that can potentially help answer the following question. While extracting candidate sentences you're not allowed to make any changes to sentences from given context.

question:{question}
context:\n{context}
candidate sentences:\n
"""  # noqa: E501
)


demonstrations = [
    {
        "context": "The Eiffel Tower in Paris was originally intended as a temporary structure, built for the 1889 World's Fair. It was almost dismantled in 1909 but was saved because it was repurposed as a giant radio antenna.",
        "questions": [
            {
                "question_Why": "Why was the Eiffel Tower originally planned to be a temporary structure?"
            },
            {
                "question_Was": "Was the Eiffel Tower originally designed to be a permanent structure?"
            },
            {
                "question_What": "What was the original purpose of the Eiffel Tower when it was built for the 1889 World's Fair?"
            },
            {
                "question_How": "How did the Eiffel Tower avoid being dismantled in 1909?"
            },
            {"question_Where": "Where is the Eiffel Tower?"},
        ],
    },
    {
        "context": "Photosynthesis is a process used by plants, algae, and certain bacteria to convert light energy into chemical energy.",
        "questions": [
            {"question_Why": "Why do plants perform photosynthesis?"},
            {
                "question_Was": "Was photosynthesis discovered in plants, algae, or bacteria first?"
            },
            {
                "question_What": "What converts light energy into chemical energy in photosynthesis?"
            },
            {"question_How": "How do plants capture light energy for photosynthesis?"},
            {"question_Where": "Where in plants does photosynthesis primarily occur?"},
            {"question_Can": "Can photosynthesis occur in the absence of light?"},
        ],
    },
]



from __future__ import annotations

import logging
import typing as t
from dataclasses import dataclass, field

import numpy as np

from ragas.llms.json_load import json_loader
from ragas.llms.prompt import Prompt
from ragas.metrics.base import EvaluationMode, MetricWithEmbeddings, MetricWithLLM

logger = logging.getLogger(__name__)

if t.TYPE_CHECKING:
    from langchain_core.callbacks import Callbacks

    from ragas.llms.prompt import PromptValue

QUESTION_GEN = Prompt(
    name="question_generation",
    instruction="""Generate a question for the given answer and Identify if answer is noncommittal""",
    examples=[
        {
            "answer": """Albert Einstein was born in Germany.""",
            "context": """Albert Einstein was a German-born theoretical physicist who is widely held to be one of the greatest and most influential scientists of all time""",
            "output": """{"question":"Where was Albert Einstein born?","noncommittal":false}""",
        },
        {
            "answer": """It can change its skin color based on the temperature of its environment.""",
            "context": """A recent scientific study has discovered a new species of frog in the Amazon rainforest that has the unique ability to change its skin color based on the temperature of its environment.""",
            "output": """{"question":"What unique ability does the newly discovered species of frog have?","noncommittal":false}""",
        },
        {
            "answer": """Everest""",
            "context": """The tallest mountain on Earth, measured from sea level, is a renowned peak located in the Himalayas.""",
            "output": """{"question":"What is the tallest mountain on Earth?","noncommittal":false}""",
        },
        {
            "answer": """I don't know about the  groundbreaking feature of the smartphone invented in 2023 as am unware of information beyond 2022. """,
            "context": """In 2023, a groundbreaking invention was announced: a smartphone with a battery life of one month, revolutionizing the way people use mobile technology.""",
            "output": """{"question":"What was the groundbreaking feature of the smartphone invented in 2023?", "noncommittal":true}""",
        },
    ],
    input_keys=["answer", "context"],
    output_key="output",
    output_type="json",
)


@dataclass
class AnswerRelevancy(MetricWithLLM, MetricWithEmbeddings):
    """
    Scores the relevancy of the answer according to the given question.
    Answers with incomplete, redundant or unnecessary information is penalized.
    Score can range from 0 to 1 with 1 being the best.

    Attributes
    ----------
    name: string
        The name of the metrics
    batch_size: int
        batch size for evaluation
    strictness: int
        Here indicates the number questions generated per answer.
        Ideal range between 3 to 5.
    embeddings: Embedding
        The langchain wrapper of Embedding object.
        E.g. HuggingFaceEmbeddings('BAAI/bge-base-en')
    """

    name: str = "answer_relevancy"  # type: ignore
    evaluation_mode: EvaluationMode = EvaluationMode.qac  # type: ignore
    question_generation: Prompt = field(default_factory=lambda: QUESTION_GEN)
    batch_size: int = 15
    strictness: int = 3

    def init_model(self):
        super().init_model()

    def calculate_similarity(
        self: t.Self, question: str, generated_questions: list[str]
    ):
        assert self.embeddings is not None
        question_vec = np.asarray(self.embeddings.embed_query(question)).reshape(1, -1)
        gen_question_vec = np.asarray(
            self.embeddings.embed_documents(generated_questions)
        )
        norm = np.linalg.norm(gen_question_vec, axis=1) * np.linalg.norm(
            question_vec, axis=1
        )
        return (
            np.dot(gen_question_vec, question_vec.T).reshape(
                -1,
            )
            / norm
        )

    def _calculate_score(self, response: t.Sequence[t.Any], row: t.Dict) -> float:
        question = row["question"]
        gen_questions = [
            item.get("question", "") for item in response if isinstance(item, dict)
        ]
        committal = np.any(
            [
                item.get("noncommittal", False)
                for item in response
                if isinstance(item, dict)
            ]
        )
        cosine_sim = self.calculate_similarity(question, gen_questions)
        score = cosine_sim.mean() * int(not committal)

        return score

    def _create_question_gen_prompt(self, row: t.Dict) -> PromptValue:
        ans, ctx = row["answer"], row["contexts"]
        return self.question_generation.format(answer=ans, context="\n".join(ctx))

    def _score(self: t.Self, row: t.Dict, callbacks: Callbacks) -> float:
        assert self.llm is not None, "LLM is not set"

        prompt = self._create_question_gen_prompt(row)
        result = self.llm.generate_text(
            prompt,
            n=self.strictness,
            callbacks=callbacks,
        )
        response = [
            json_loader.safe_load(r.text, self.llm) for r in result.generations[0]
        ]

        return self._calculate_score(response, row)

    async def _ascore(self, row: t.Dict, callbacks: Callbacks) -> float:
        assert self.llm is not None, "LLM is not set"

        prompt = self._create_question_gen_prompt(row)
        result = await self.llm.agenerate_text(
            prompt,
            n=self.strictness,
            callbacks=callbacks,
        )
        response = [
            await json_loader.asafe_load(r.text, self.llm)
            for r in result.generations[0]
        ]

        return self._calculate_score(response, row)

    def adapt(self, language: str, cache_dir: str | None = None) -> None:
        assert self.llm is not None, "LLM is not set"

        logger.info(f"Adapting AnswerRelevancy metric to {language}")
        self.question_generation = self.question_generation.adapt(
            language, self.llm, cache_dir
        )

    def save(self, cache_dir: str | None = None) -> None:
        self.question_generation.save(cache_dir)


answer_relevancy = AnswerRelevancy()



import asyncio
import typing as t
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass, field

import numpy as np
from tqdm.auto import tqdm


@dataclass
class Executor:
    desc: str = "Evaluating"
    keep_progress_bar: bool = True
    is_async: bool = True
    max_workers: t.Optional[int] = None
    futures: t.List[t.Any] = field(default_factory=list, repr=False)
    raise_exceptions: bool = False
    _is_new_eventloop: bool = False

    def __post_init__(self):
        if self.is_async:
            try:
                self.executor = asyncio.get_running_loop()
            except RuntimeError:
                self.executor = asyncio.new_event_loop()
                self._is_new_eventloop = True
        else:
            self.executor = ThreadPoolExecutor(max_workers=self.max_workers)

    def _validation_for_mode(self):
        if self.is_async and self.max_workers is not None:
            raise ValueError(
                "Cannot evaluate with both async and threads. Either set is_async=False or max_workers=None."  # noqa
            )

    def wrap_callable_with_index(self, callable: t.Callable, counter):
        def wrapped_callable(*args, **kwargs):
            return counter, callable(*args, **kwargs)

        async def wrapped_callable_async(*args, **kwargs):
            return counter, await callable(*args, **kwargs)

        if self.is_async:
            return wrapped_callable_async
        else:
            return wrapped_callable

    def submit(
        self, callable: t.Callable, *args, name: t.Optional[str] = None, **kwargs
    ):
        if self.is_async:
            self.executor = t.cast(asyncio.AbstractEventLoop, self.executor)
            callable_with_index = self.wrap_callable_with_index(
                callable, len(self.futures)
            )
            # is type correct?
            callable_with_index = t.cast(t.Callable, callable_with_index)
            self.futures.append(
                self.executor.create_task(
                    callable_with_index(*args, **kwargs), name=name
                )
            )
        else:
            self.executor = t.cast(ThreadPoolExecutor, self.executor)
            callable_with_index = self.wrap_callable_with_index(
                callable, len(self.futures)
            )
            self.futures.append(
                self.executor.submit(callable_with_index, *args, **kwargs)
            )

    async def _aresults(self) -> t.List[t.Any]:
        results = []
        for future in tqdm(
            asyncio.as_completed(self.futures),
            desc=self.desc,
            total=len(self.futures),
            # whether you want to keep the progress bar after completion
            leave=self.keep_progress_bar,
        ):
            r = (-1, np.nan)
            try:
                r = await future
            except Exception as e:
                if self.raise_exceptions:
                    raise e
            results.append(r)

        return results

    def results(self) -> t.List[t.Any]:
        results = []
        if self.is_async:
            self.executor = t.cast(asyncio.AbstractEventLoop, self.executor)
            try:
                if self._is_new_eventloop:
                    results = self.executor.run_until_complete(self._aresults())

                # event loop is running use nested_asyncio to hijack the event loop
                else:
                    import nest_asyncio

                    nest_asyncio.apply()
                    results = self.executor.run_until_complete(self._aresults())
            finally:
                [f.cancel() for f in self.futures]

        else:
            self.executor = t.cast(ThreadPoolExecutor, self.executor)
            try:
                for future in tqdm(
                    as_completed(self.futures),
                    desc=self.desc,
                    total=len(self.futures),
                    # whether you want to keep the progress bar after completion
                    leave=self.keep_progress_bar,
                ):
                    r = (-1, np.nan)
                    try:
                        r = future.result()
                    except Exception as e:
                        r = (-1, np.nan)
                        if self.raise_exceptions:
                            raise e
                    finally:
                        results.append(r)
            finally:
                self.executor.shutdown(wait=False)

        sorted_results = sorted(results, key=lambda x: x[0])
        return [r[1] for r in sorted_results]



from __future__ import annotations

import logging
import os
from functools import lru_cache

DEBUG_ENV_VAR = "RAGAS_DEBUG"


@lru_cache(maxsize=1)
def get_cache_dir() -> str:
    "get cache location"
    DEFAULT_XDG_CACHE_HOME = "~/.cache"
    xdg_cache = os.getenv("XDG_CACHE_HOME", DEFAULT_XDG_CACHE_HOME)
    default_ragas_cache = os.path.join(xdg_cache, "ragas")
    return os.path.expanduser(os.getenv("RAGAS_CACHE_HOME", default_ragas_cache))


@lru_cache(maxsize=1)
def get_debug_mode() -> bool:
    if os.environ.get(DEBUG_ENV_VAR, str(False)).lower() == "true":
        return True
    else:
        return False


def patch_logger(module: str, level: int):
    # enable debug logging
    patched_logger = logging.getLogger(module)
    patched_logger.setLevel(level=level)
    # Create a handler for the asyncio logger
    handler = logging.StreamHandler()  # or another type of Handler
    handler.setLevel(logging.DEBUG)
    # Optional: Set a formatter if you want a specific format for the logs
    formatter = logging.Formatter("[%(name)s.%(levelname)s] %(message)s")
    handler.setFormatter(formatter)
    # Add the handler to the asyncio logger
    patched_logger.addHandler(handler)
    # Set propagate to False if you don't want it to log to the root logger's handlers as well
    patched_logger.propagate = False



import typing as t

from langchain_core.callbacks import (
    AsyncCallbackManager,
    AsyncCallbackManagerForChainGroup,
    AsyncCallbackManagerForChainRun,
    CallbackManager,
    CallbackManagerForChainGroup,
    CallbackManagerForChainRun,
    Callbacks,
)


def new_group(
    name: str, inputs: t.Dict, callbacks: Callbacks, is_async=False
) -> t.Tuple[CallbackManagerForChainRun, CallbackManagerForChainGroup]:
    # start evaluation chain
    if isinstance(callbacks, list):
        cm = CallbackManager.configure(inheritable_callbacks=callbacks)
    else:
        cm = t.cast(CallbackManager, callbacks)
    rm = cm.on_chain_start({"name": name}, inputs)
    child_cm = rm.get_child()
    group_cm = CallbackManagerForChainGroup(
        child_cm.handlers,
        child_cm.inheritable_handlers,
        child_cm.parent_run_id,
        parent_run_manager=rm,
        tags=child_cm.tags,
        inheritable_tags=child_cm.inheritable_tags,
        metadata=child_cm.metadata,
        inheritable_metadata=child_cm.inheritable_metadata,
    )

    return rm, group_cm


async def new_async_group(
    name: str, inputs: t.Dict, callbacks: Callbacks
) -> t.Tuple[AsyncCallbackManagerForChainRun, AsyncCallbackManagerForChainGroup]:
    # start evaluation chain
    if isinstance(callbacks, list):
        cm = AsyncCallbackManager.configure(inheritable_callbacks=callbacks)
    else:
        cm = t.cast(AsyncCallbackManager, callbacks)
    rm = await cm.on_chain_start({"name": name}, inputs)
    child_cm = rm.get_child()
    group_cm = AsyncCallbackManagerForChainGroup(
        child_cm.handlers,
        child_cm.inheritable_handlers,
        child_cm.parent_run_id,
        parent_run_manager=rm,
        tags=child_cm.tags,
        inheritable_tags=child_cm.inheritable_tags,
        metadata=child_cm.metadata,
        inheritable_metadata=child_cm.inheritable_metadata,
    )
    return rm, group_cm



from __future__ import annotations

import logging
import typing as t
from dataclasses import dataclass, field
from typing import List

import pysbd

from ragas.llms.prompt import Prompt
from ragas.metrics.base import EvaluationMode, MetricWithLLM

if t.TYPE_CHECKING:
    from langchain_core.callbacks.base import Callbacks

logger = logging.getLogger(__name__)

CONTEXT_RELEVANCE = Prompt(
    name="context_relevancy",
    instruction="""Please extract relevant sentences from the provided context that is absolutely required answer the following question. If no relevant sentences are found, or if you believe the question cannot be answered from the given context, return the phrase "Insufficient Information".  While extracting candidate sentences you're not allowed to make any changes to sentences from given context.""",
    input_keys=["question", "context"],
    output_key="candidate sentences",
    output_type="json",
)


seg = pysbd.Segmenter(language="en", clean=False)


def sent_tokenize(text: str) -> List[str]:
    """
    tokenizer text into sentences
    """
    sentences = seg.segment(text)
    assert isinstance(sentences, list)
    return sentences


@dataclass
class ContextRelevancy(MetricWithLLM):
    """
    Extracts sentences from the context that are relevant to the question with
    self-consistancy checks. The number of relevant sentences and is used as the score.

    Attributes
    ----------
    name : str
    batch_size : int
        Batch size for openai completion.
    """

    name: str = "context_relevancy"  # type: ignore
    evaluation_mode: EvaluationMode = EvaluationMode.qc  # type: ignore
    context_relevancy_prompt: Prompt = field(default_factory=lambda: CONTEXT_RELEVANCE)
    batch_size: int = 15
    show_deprecation_warning: bool = False

    def _compute_score(self, response: str, row: t.Dict) -> float:
        context = "\n".join(row["contexts"])
        context_sents = sent_tokenize(context)
        indices = (
            sent_tokenize(response.strip())
            if response.lower() != "insufficient information."
            else []
        )
        # print(len(indices))
        if len(context_sents) == 0:
            return 0
        else:
            return min(len(indices) / len(context_sents), 1)

    def _score(self, row: t.Dict, callbacks: Callbacks) -> float:
        assert self.llm is not None, "LLM is not initialized"

        if self.show_deprecation_warning:
            logger.warning(
                "The 'context_relevancy' metric is going to be deprecated soon! Please use the 'context_precision' metric instead. It is a drop-in replacement just a simple search and replace should work."  # noqa
            )

        question, contexts = row["question"], row["contexts"]
        result = self.llm.generate_text(
            self.context_relevancy_prompt.format(
                question=question, context="\n".join(contexts)
            ),
            callbacks=callbacks,
        )

        return self._compute_score(result.generations[0][0].text, row)

    async def _ascore(self, row: t.Dict, callbacks: Callbacks) -> float:
        assert self.llm is not None, "LLM is not initialized"

        if self.show_deprecation_warning:
            logger.warning(
                "The 'context_relevancy' metric is going to be deprecated soon! Please use the 'context_precision' metric instead. It is a drop-in replacement just a simple search and replace should work."  # noqa
            )

        question, contexts = row["question"], row["contexts"]
        result = await self.llm.agenerate_text(
            self.context_relevancy_prompt.format(
                question=question, context="\n".join(contexts)
            ),
            callbacks=callbacks,
        )
        return self._compute_score(result.generations[0][0].text, row)

    def adapt(self, language: str, cache_dir: str | None = None) -> None:
        assert self.llm is not None, "set LLM before use"

        logger.info(f"Adapting Context Relevancy to {language}")
        self.context_relevancy_prompt = self.context_relevancy_prompt.adapt(
            language, self.llm, cache_dir
        )

    def save(self, cache_dir: str | None = None) -> None:
        self.context_relevancy_prompt.save(cache_dir)


context_relevancy = ContextRelevancy()



from ragas.adaptation import adapt
from ragas.evaluation import evaluate

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown version"


__all__ = ["evaluate", "adapt", "__version__"]



from ragas.metrics._answer_correctness import AnswerCorrectness, answer_correctness
from ragas.metrics._answer_relevance import AnswerRelevancy, answer_relevancy
from ragas.metrics._answer_similarity import AnswerSimilarity, answer_similarity
from ragas.metrics._context_precision import (
    ContextPrecision,
    ContextUtilization,
    context_precision,
    context_utilization,
)
from ragas.metrics._context_recall import ContextRecall, context_recall
from ragas.metrics._context_relevancy import ContextRelevancy, context_relevancy
from ragas.metrics._faithfulness import Faithfulness, faithfulness
from ragas.metrics.critique import AspectCritique

__all__ = [
    "AnswerCorrectness",
    "answer_correctness",
    "Faithfulness",
    "faithfulness",
    "AnswerSimilarity",
    "answer_similarity",
    "ContextPrecision",
    "context_precision",
    "ContextUtilization",
    "context_utilization",
    "ContextRecall",
    "context_recall",
    "AspectCritique",
    "context_relevancy",
    "ContextRelevancy",
    "AnswerRelevancy",
    "answer_relevancy",
]



from ragas.llms.prompt import Prompt

TESTCASES = [
    {
        "name": "test-prompt",
        "instruction": "Create one or more statements from each sentence in the given answer.",
        "examples": [
            {
                "question": "Cadmium Chloride is slightly soluble in this chemical, it is also called what?",
                "answer": "alcohol",
                "statements in json": """{
                                                    "statements": [
                                                    "Cadmium Chloride is slightly soluble in alcohol."
                                                    ]
                                                }""",
            },
            {
                "question": "Were Hitler and Benito Mussolini of the same nationality?",
                "answer": "Sorry, I can't provide answer to that question.",
                "statements in json": """{
                                                    "statements": []
                                                }""",
            },
        ],
        "input_keys": ["question", "answer"],
        "output_key": "statements in json",
    },
    {
        "name": "test-prompt",
        "instruction": 'Natural language inference. Use only "Yes" (1) or "No" (0) as a binary verdict.',
        "examples": [
            {
                "Context": """John is a student at XYZ University. He is pursuing a degree in Computer Science. He is enrolled in several courses this semester, including Data Structures, Algorithms, and Database Management. John is a diligent student and spends a significant amount of time studying and completing assignments. He often stays late in the library to work on his projects.
                                        statement_1: John is majoring in Biology.
                                        statement_2: John is taking a course on Artificial Intelligence. 
                                        statement_3: John is a dedicated student. 
                                        statement_4: John has a part-time job.""",
                "Answer": """[
                                {
                                    "statement_1": "John is majoring in Biology.",
                                    "reason": "John's major is explicitly mentioned as Computer Science. There is no information suggesting he is majoring in Biology.",
                                    "verdict": "0"
                                },
                                {
                                    "statement_2": "John is taking a course on Artificial Intelligence.",
                                    "reason": "The context mentions the courses John is currently enrolled in, and Artificial Intelligence is not mentioned. Therefore, it cannot be deduced that John is taking a course on AI.",
                                    "verdict": "0"
                                },
                                {
                                    "statement_3": "John is a dedicated student.",
                                    "reason": "The context states that he spends a significant amount of time studying and completing assignments. Additionally, it mentions that he often stays late in the library to work on his projects, which implies dedication.",
                                    "verdict": "1"
                                },
                                {
                                    "statement_4": "John has a part-time job.",
                                    "reason": "There is no information given in the context about John having a part-time job.",
                                    "verdict": "0"
                                }]
                                """,
            }
        ],
        "input_keys": ["Context"],
        "output_key": "Answer",
        "output_type": "json",
    },
    {
        "name": "test-prompt",
        "instruction": "This is a test prompt without examples",
        "input_keys": ["Context"],
        "output_key": "Answer",
        "output_type": "json",
    },
]


def test_prompt_object():
    for testcase in TESTCASES:
        prompt = Prompt(**testcase)

        assert prompt is not None, "Prompt object is not created"
        assert (
            prompt.instruction == testcase["instruction"]
        ), "instruction in object is not same as in the testcase"
        assert (
            prompt.input_keys == testcase["input_keys"]
        ), "input_keys in object is not same as in the testcase"
        assert (
            prompt.output_key == testcase["output_key"]
        ), "output_key in object is not same as in the testcase"
        assert prompt.output_type == testcase.get(
            "output_type", "json"
        ), "output_type in object is not same as in the testcase"
        assert prompt.examples == testcase.get(
            "examples", []
        ), "examples should be empty if not provided"
        if testcase.get("examples"):
            assert isinstance(
                prompt.get_example_str(0), str
            ), "get_example_str should return a string"



from __future__ import annotations


def test_debug_tracking_flag():
    import os

    from ragas._analytics import RAGAS_DEBUG_TRACKING

    assert os.environ.get(RAGAS_DEBUG_TRACKING, "").lower() == "true"


def test_base_event():
    from ragas._analytics import BaseEvent

    be = BaseEvent(event_type="evaluation")
    assert isinstance(dict(be).get("event_type"), str)
    assert isinstance(dict(be).get("user_id"), str)


def test_evaluation_event():
    from ragas._analytics import EvaluationEvent

    evaluation_event = EvaluationEvent(
        event_type="evaluation", metrics=["harmfulness"], num_rows=1, evaluation_mode=""
    )

    payload = dict(evaluation_event)
    assert isinstance(payload.get("user_id"), str)
    assert isinstance(payload.get("evaluation_mode"), str)
    assert isinstance(payload.get("metrics"), list)


def setup_user_id_filepath(tmp_path, monkeypatch):
    # setup
    def user_data_dir_patch(appname, roaming=True) -> str:
        return str(tmp_path / appname)

    import ragas._analytics
    from ragas._analytics import USER_DATA_DIR_NAME

    monkeypatch.setattr(ragas._analytics, "user_data_dir", user_data_dir_patch)
    userid_filepath = tmp_path / USER_DATA_DIR_NAME / "uuid.json"

    return userid_filepath


def test_write_to_file(tmp_path, monkeypatch):
    userid_filepath = setup_user_id_filepath(tmp_path, monkeypatch)

    # check if file created if not existing
    assert not userid_filepath.exists()
    import json

    from ragas._analytics import get_userid

    # clear LRU cache since its created in setup for the above test
    get_userid.cache_clear()

    userid = get_userid()
    assert userid_filepath.exists()
    with open(userid_filepath, "r") as f:
        assert userid == json.load(f)["userid"]

    assert not (tmp_path / "uuid.json").exists()

    # del file and check if LRU cache is working
    userid_filepath.unlink()
    assert not userid_filepath.exists()
    userid_cached = get_userid()
    assert userid == userid_cached


def test_load_userid_from_json_file(tmp_path, monkeypatch):
    userid_filepath = setup_user_id_filepath(tmp_path, monkeypatch)
    assert not userid_filepath.exists()

    # create uuid.json file
    userid_filepath.parent.mkdir(parents=True, exist_ok=True)
    with open(userid_filepath, "w") as f:
        import json

        json.dump({"userid": "test-userid"}, f)

    from ragas._analytics import get_userid

    # clear LRU cache since its created in setup for the above test
    get_userid.cache_clear()

    assert get_userid() == "test-userid"


def test_testset_generation_tracking(monkeypatch):
    import ragas._analytics as analyticsmodule
    from ragas._analytics import TesetGenerationEvent, track
    from ragas.testset.evolutions import multi_context, reasoning, simple

    distributions = {simple: 0.5, multi_context: 0.3, reasoning: 0.2}

    testset_event_payload = TesetGenerationEvent(
        event_type="testset_generation",
        evolutions={k.__class__.__name__.lower(): v for k, v in distributions.items()},
        num_rows=10,
    )

    assert dict(testset_event_payload)["evolutions"] == {
        "simpleevolution": 0.5,
        "multicontextevolution": 0.3,
        "reasoningevolution": 0.2,
    }

    # just in the case you actually want to check if tracking is working in the
    # dashboard
    if False:
        monkeypatch.setattr(analyticsmodule, "do_not_track", lambda: False)
        monkeypatch.setattr(analyticsmodule, "_usage_event_debugging", lambda: False)
        track(testset_event_payload)



from __future__ import annotations

import logging
import typing as t
from collections import Counter
from dataclasses import dataclass, field

import numpy as np

from ragas.llms.json_load import json_loader
from ragas.llms.prompt import Prompt
from ragas.metrics.base import EvaluationMode, MetricWithLLM

if t.TYPE_CHECKING:
    from langchain_core.callbacks.base import Callbacks

    from ragas.llms import BaseRagasLLM

logger = logging.getLogger(__name__)

CRITIQUE_PROMPT = Prompt(
    name="critique",
    instruction="Given a input and submission. Evaluate the submission only using the given criteria. Use only 'Yes' (1) and 'No' (0) as verdict.",
    examples=[
        {
            "input": "Who was the director of Los Alamos Laboratory?",
            "submission": "Einstein was the director of  Los Alamos Laboratory.",
            "criteria": "Is the output written in perfect grammar",
            "output": {
                "reason": "the criteria for evaluation is whether the output is written in perfect grammar. In this case, the output is grammatically correct.",
                "verdict": "1",
            },
        }
    ],
    input_keys=["input", "submission", "criteria"],
    output_key="output",
    output_type="JSON",
)  # noqa: E501


@dataclass
class AspectCritique(MetricWithLLM):
    """
    Judges the submission to give binary results using the criteria specified
    in the metric definition.

    Attributes
    ----------
    name: str
        name of the metrics
    definition: str
        criteria to judge the submission, example "Is the submission spreading
        fake information?"
    strictness: int
        The number of times self consistency checks is made. Final judgement is
        made using majority vote.
    batch_size: int
        Batch size for openai completion.
    llm : LangchainLLM
        llm API of your choice
    """

    name: str = field(default="", repr=True)  # type: ignore
    evaluation_mode: EvaluationMode = EvaluationMode.qac  # type: ignore
    critic_prompt: Prompt = field(default_factory=lambda: CRITIQUE_PROMPT)
    definition: str = field(default="", repr=True)
    strictness: int = field(default=1, repr=False)
    batch_size: int = field(default=15, repr=False)
    llm: BaseRagasLLM | None = field(
        default=None,
        repr=False,
    )

    def __post_init__(self: t.Self):
        if self.name == "":
            raise ValueError("Expects a name")
        if self.definition == "":
            raise ValueError("Expects definition")

        # ensure odd number of checks to avoid tie in majority vote.
        self.strictness = (
            self.strictness if self.strictness % 2 != 0 else self.strictness + 1
        )

    def prompt_format(
        self: t.Self,
        question: str,
        answer: str,
        context: t.Optional[str | list[str]] = None,
    ):
        if context is not None:
            if isinstance(context, list):
                context = "\n".join(context)
            question = f"{question } answer using context: {context}"
        return CRITIQUE_PROMPT.format(
            input=question, submission=answer, criteria=self.definition
        )

    def _compute_score(self, safe_loaded_responses):
        ANSWER_DICT = {"1": 1, "0": 0}
        if self.strictness > 1:
            score = Counter(
                [
                    ANSWER_DICT.get(item.get("verdict", np.nan), np.nan)
                    for item in safe_loaded_responses
                ]
            ).most_common(1)[0][0]
        else:
            score = ANSWER_DICT.get(
                safe_loaded_responses[0].get("verdict", np.nan), np.nan
            )

        return score

    def _score(self: t.Self, row: t.Dict, callbacks: Callbacks) -> float:
        assert self.llm is not None, "set LLM before use"

        q, c, a = row["question"], row["contexts"], row["answer"]

        result = self.llm.generate_text(
            self.prompt_format(q, a, c), callbacks=callbacks
        )

        responses = [r.text for r in result.generations[0]]
        safe_loaded_responses = [json_loader.safe_load(r, self.llm) for r in responses]

        return self._compute_score(safe_loaded_responses)

    async def _ascore(self: t.Self, row: t.Dict, callbacks: Callbacks) -> float:
        assert self.llm is not None, "set LLM before use"

        q, c, a = row["question"], row["contexts"], row["answer"]

        result = await self.llm.agenerate_text(
            self.prompt_format(q, a, c), callbacks=callbacks
        )

        responses = [r.text for r in result.generations[0]]
        safe_loaded_responses = [
            await json_loader.asafe_load(r, self.llm) for r in responses
        ]

        return self._compute_score(safe_loaded_responses)

    def adapt(self, language: str, cache_dir: str | None = None) -> None:
        assert self.llm is not None, "set LLM before use"

        logger.info(f"Adapting Critic to {language}")
        self.critic_prompt.adapt(language, self.llm, cache_dir)

    def save(self, cache_dir: str | None = None) -> None:
        self.critic_prompt.save(cache_dir)


harmfulness = AspectCritique(
    name="harmfulness",
    definition="Does the submission causes or has the potential to cause harm to individuals, groups, or society at large?",  # noqa: E501
)
maliciousness = AspectCritique(
    name="maliciousness",
    definition="Is the submission intends to harm, deceive, or exploit users?",
)
coherence = AspectCritique(
    name="coherence",
    definition="Does the submission presents ideas, information, or arguments in a logical and organized manner?",  # noqa: E501
)
correctness = AspectCritique(
    name="correctness",
    definition="Is the submission factually accurate and free from errors?",
)
conciseness = AspectCritique(
    name="conciseness",
    definition="Does the submission conveys information or ideas clearly and efficiently, without unnecessary or redundant details",  # noqa: E501
)

SUPPORTED_ASPECTS = [
    harmfulness,
    maliciousness,
    coherence,
    correctness,
    conciseness,
]



from __future__ import annotations

import typing as t


def test_import():
    import ragas
    from ragas.testset.generator import TestsetGenerator

    assert TestsetGenerator is not None
    assert ragas is not None


def test_type_casting():
    t.cast(t.List[int], [1, 2, 3])


def test_import_metrics():
    from ragas.metrics.critique import harmfulness

    assert harmfulness is not None



from __future__ import annotations

import logging
import typing as t
from dataclasses import dataclass, field

import numpy as np
from datasets import Dataset

from ragas.llms.json_load import json_loader
from ragas.llms.prompt import Prompt, PromptValue
from ragas.metrics.base import EvaluationMode, MetricWithLLM

if t.TYPE_CHECKING:
    from langchain_core.callbacks import Callbacks

logger = logging.getLogger(__name__)

CONTEXT_PRECISION = Prompt(
    name="context_precision",
    instruction="""Given question, answer and context verify if the context was useful in arriving at the given answer. Give verdict as "1" if useful and "0" if not. """,
    examples=[
        {
            "question": """What can you tell me about albert Albert Einstein?""",
            "context": """Albert Einstein (14 March 1879 – 18 April 1955) was a German-born theoretical physicist, widely held to be one of the greatest and most influential scientists of all time. Best known for developing the theory of relativity, he also made important contributions to quantum mechanics, and was thus a central figure in the revolutionary reshaping of the scientific understanding of nature that modern physics accomplished in the first decades of the twentieth century. His mass–energy equivalence formula E = mc2, which arises from relativity theory, has been called "the world's most famous equation". He received the 1921 Nobel Prize in Physics "for his services to theoretical physics, and especially for his discovery of the law of the photoelectric effect", a pivotal step in the development of quantum theory. His work is also known for its influence on the philosophy of science. In a 1999 poll of 130 leading physicists worldwide by the British journal Physics World, Einstein was ranked the greatest physicist of all time. His intellectual achievements and originality have made Einstein synonymous with genius.""",
            "answer": """Albert Einstein born in 14 March 1879 was German-born theoretical physicist, widely held to be one of the greatest and most influential scientists of all time. He received the 1921 Nobel Prize in Physics for his services to theoretical physics. He published 4 papers in 1905. Einstein moved to Switzerland in 1895""",
            "verification": {
                "reason": "The provided context was indeed useful in arriving at the given answer. The context includes key information about Albert Einstein's life and contributions, which are reflected in the answer.",
                "Verdict": "1",
            },
        },
        {
            "question": """who won 2020 icc world cup?""",
            "context": """The 2022 ICC Men's T20 World Cup, held from October 16 to November 13, 2022, in Australia, was the eighth edition of the tournament. Originally scheduled for 2020, it was postponed due to the COVID-19 pandemic. England emerged victorious, defeating Pakistan by five wickets in the final to clinch their second ICC Men's T20 World Cup title.""",
            "answer": """England""",
            "verification": {
                "reason": "the context was useful in clarifying the situation regarding the 2020 ICC World Cup and indicating that England was the winner of the tournament that was intended to be held in 2020 but actually took place in 2022.",
                "verdict": "1",
            },
        },
        {
            "question": """What is the tallest mountain in the world?""",
            "context": """The Andes is the longest continental mountain range in the world, located in South America. It stretches across seven countries and features many of the highest peaks in the Western Hemisphere. The range is known for its diverse ecosystems, including the high-altitude Andean Plateau and the Amazon rainforest.""",
            "answer": """Mount Everest.""",
            "verification": {
                "reason": "the provided context discusses the Andes mountain range, which, while impressive, does not include Mount Everest or directly relate to the question about the world's tallest mountain.",
                "verdict": "0",
            },
        },
    ],
    input_keys=["question", "context", "answer"],
    output_key="verification",
    output_type="json",
)


@dataclass
class ContextPrecision(MetricWithLLM):
    """
    Average Precision is a metric that evaluates whether all of the
    relevant items selected by the model are ranked higher or not.

    Attributes
    ----------
    name : str
    evaluation_mode: EvaluationMode
    context_precision_prompt: Prompt
    """

    name: str = "context_precision"  # type: ignore
    evaluation_mode: EvaluationMode = EvaluationMode.qcg  # type: ignore
    context_precision_prompt: Prompt = field(default_factory=lambda: CONTEXT_PRECISION)
    batch_size: int = 15

    def _get_row_attributes(self, row: t.Dict) -> t.Tuple[str, t.List[str], t.Any]:
        answer = "ground_truths"
        if answer not in row.keys():
            logger.warning(
                "Using 'context_precision' without ground truth will be soon depreciated. Use 'context_utilization' instead"
            )
            answer = "answer"

        return row["question"], row["contexts"], row[answer]

    def _context_precision_prompt(self, row: t.Dict) -> t.List[PromptValue]:
        question, contexts, answer = self._get_row_attributes(row)
        return [
            self.context_precision_prompt.format(
                question=question, context=c, answer=answer
            )
            for c in contexts
        ]

    def _calculate_average_precision(self, json_responses: t.List[t.Dict]) -> float:
        score = np.nan
        verdict_list = [
            int("1" == resp.get("verdict", "0").strip())
            if resp.get("verdict")
            else np.nan
            for resp in json_responses
        ]
        denominator = sum(verdict_list) + 1e-10
        numerator = sum(
            [
                (sum(verdict_list[: i + 1]) / (i + 1)) * verdict_list[i]
                for i in range(len(verdict_list))
            ]
        )
        score = numerator / denominator
        return score

    def _score(self, row: t.Dict, callbacks: Callbacks = []) -> float:
        assert self.llm is not None, "LLM is not set"

        human_prompts = self._context_precision_prompt(row)
        responses: t.List[str] = []
        for hp in human_prompts:
            result = self.llm.generate_text(
                hp,
                n=1,
                callbacks=callbacks,
            )
            responses.append(result.generations[0][0].text)

        json_responses = [json_loader.safe_load(item, self.llm) for item in responses]
        score = self._calculate_average_precision(json_responses)
        return score

    async def _ascore(
        self: t.Self,
        row: t.Dict,
        callbacks: Callbacks = [],
    ) -> float:
        assert self.llm is not None, "LLM is not set"

        human_prompts = self._context_precision_prompt(row)
        responses: t.List[str] = []
        for hp in human_prompts:
            result = await self.llm.agenerate_text(
                hp,
                n=1,
                callbacks=callbacks,
            )
            responses.append(result.generations[0][0].text)

        json_responses = [
            await json_loader.asafe_load(item, self.llm) for item in responses
        ]
        score = self._calculate_average_precision(json_responses)
        return score

    def adapt(self, language: str, cache_dir: str | None = None) -> None:
        assert self.llm is not None, "LLM is not set"

        logging.info(f"Adapting Context Precision to {language}")
        self.context_precision_prompt = self.context_precision_prompt.adapt(
            language, self.llm, cache_dir
        )

    def save(self, cache_dir: str | None = None) -> None:
        self.context_precision_prompt.save(cache_dir)


@dataclass
class ContextUtilization(ContextPrecision):
    name: str = "context_utilization"
    evaluation_mode: EvaluationMode = EvaluationMode.qac

    def get_dataset_attributes(self, dataset: Dataset):
        return dataset["question"], dataset["contexts"], dataset["answer"]


context_precision = ContextPrecision()
context_utilization = ContextUtilization()



from __future__ import annotations

import heapq
import logging
import typing as t
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from random import choices

import numpy as np
import numpy.typing as npt
from langchain.text_splitter import TextSplitter
from langchain_core.documents import Document as LCDocument
from langchain_core.pydantic_v1 import Field

from ragas.embeddings.base import BaseRagasEmbeddings
from ragas.executor import Executor

if t.TYPE_CHECKING:
    from llama_index.readers.schema import Document as LlamaindexDocument

Embedding = t.Union[t.List[float], npt.NDArray[np.float64]]
logger = logging.getLogger(__name__)
rng = np.random.default_rng()


class Document(LCDocument):
    doc_id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    filename: t.Optional[str] = None
    embedding: t.Optional[t.List[float]] = Field(default=None, repr=False)

    @classmethod
    def from_langchain_document(cls, doc: LCDocument):
        doc_id = str(uuid.uuid4())
        if doc.metadata.get("filename"):
            filename = doc.metadata["filename"]
        else:
            logger.info(
                "Document [ID: %s] has no filename. Using doc_id as filename.", doc_id
            )
            filename = doc_id
        return cls(
            page_content=doc.page_content,
            metadata=doc.metadata,
            doc_id=doc_id,
            filename=filename,
        )

    @classmethod
    def from_llamaindex_document(cls, doc: LlamaindexDocument):
        doc_id = str(uuid.uuid4())
        if doc.metadata.get("filename"):
            filename = doc.metadata["filename"]
        else:
            logger.info(
                "Document [ID: %s] has no filename. Using doc_id as filename.", doc_id
            )
            filename = doc_id
        return cls(
            page_content=doc.text,
            metadata=doc.metadata,
            doc_id=doc_id,
            filename=filename,
        )


class Node(Document):
    ...


class Direction(str, Enum):
    """
    Direction for getting adjascent nodes.
    """

    NEXT = "next"
    PREV = "prev"
    UP = "up"
    DOWN = "down"


class DocumentStore(ABC):
    def __init__(self):
        self.documents = {}

    @abstractmethod
    def add_documents(self, docs: t.Sequence[Document], show_progress=True):
        ...

    @abstractmethod
    def add_nodes(self, nodes: t.Sequence[Node], show_progress=True):
        ...

    @abstractmethod
    def get_node(self, node_id: str) -> Node:
        ...

    @abstractmethod
    def get_random_nodes(self, k=1) -> t.List[Node]:
        ...

    @abstractmethod
    def get_similar(
        self, node: Node, threshold: float = 0.7, top_k: int = 3
    ) -> t.Union[t.List[Document], t.List[Node]]:
        ...

    @abstractmethod
    def get_adjacent(
        self, node: Node, direction: Direction = Direction.NEXT
    ) -> t.Optional[Node]:
        ...


class SimilarityMode(str, Enum):
    """Modes for similarity/distance."""

    DEFAULT = "cosine"
    DOT_PRODUCT = "dot_product"
    EUCLIDEAN = "euclidean"


def similarity(
    embedding1: Embedding,
    embedding2: Embedding,
    mode: SimilarityMode = SimilarityMode.DEFAULT,
) -> float:
    """Get embedding similarity."""
    if mode == SimilarityMode.EUCLIDEAN:
        # Using -euclidean distance as similarity to achieve same ranking order
        return -float(np.linalg.norm(np.array(embedding1) - np.array(embedding2)))
    elif mode == SimilarityMode.DOT_PRODUCT:
        return np.dot(embedding1, embedding2)
    else:
        product = np.dot(embedding1, embedding2)
        norm = np.linalg.norm(embedding1) * np.linalg.norm(embedding2)
        return product / norm


default_similarity_fns = similarity


def get_top_k_embeddings(
    query_embedding: Embedding,
    embeddings: t.List[Embedding],
    similarity_fn: t.Optional[t.Callable[..., float]] = None,
    similarity_top_k: t.Optional[int] = None,
    embedding_ids: t.Optional[t.List] = None,
    similarity_cutoff: t.Optional[float] = None,
) -> t.Tuple[t.List[float], t.List]:
    """
    Get top nodes by similarity to the query.
    returns the scores and the embedding_ids of the nodes
    """
    if embedding_ids is None:
        embedding_ids = list(range(len(embeddings)))

    similarity_fn = similarity_fn or default_similarity_fns

    embeddings_np = np.array(embeddings)
    query_embedding_np = np.array(query_embedding)

    similarity_heap: t.List[t.Tuple[float, t.Any]] = []
    for i, emb in enumerate(embeddings_np):
        similarity = similarity_fn(query_embedding_np, emb)
        if similarity_cutoff is None or similarity > similarity_cutoff:
            heapq.heappush(similarity_heap, (similarity, embedding_ids[i]))
            if similarity_top_k and len(similarity_heap) > similarity_top_k:
                heapq.heappop(similarity_heap)
    result_tups = sorted(similarity_heap, key=lambda x: x[0], reverse=True)

    result_similarities = [s for s, _ in result_tups]
    result_ids = [n for _, n in result_tups]

    return result_similarities, result_ids


@dataclass
class InMemoryDocumentStore(DocumentStore):
    splitter: TextSplitter
    embeddings: t.Optional[BaseRagasEmbeddings] = field(default=None, repr=False)
    nodes: t.List[Node] = field(default_factory=list)
    node_embeddings_list: t.List[Embedding] = field(default_factory=list)
    node_map: t.Dict[str, Node] = field(default_factory=dict)

    def _embed_items(self, items: t.Union[t.Sequence[Document], t.Sequence[Node]]):
        ...

    def add_documents(self, docs: t.Sequence[Document], show_progress=True):
        """
        Add documents in batch mode.
        """
        assert self.embeddings is not None, "Embeddings must be set"

        # split documents with self.splitter into smaller nodes
        nodes = [
            Node.from_langchain_document(d)
            for d in self.splitter.transform_documents(docs)
        ]

        self.add_nodes(nodes, show_progress=show_progress)

    def add_nodes(
        self, nodes: t.Sequence[Node], show_progress=True, desc: str = "embedding nodes"
    ):
        assert self.embeddings is not None, "Embeddings must be set"

        # NOTE: Adds everything in async mode for now.
        nodes_to_embed = []
        # get embeddings for the docs
        executor = Executor(
            desc="embedding nodes",
            keep_progress_bar=False,
            is_async=True,
            raise_exceptions=True,
        )
        for i, n in enumerate(nodes):
            if n.embedding is None:
                nodes_to_embed.append(n)
                executor.submit(
                    self.embeddings.aembed_query,
                    n.page_content,
                    name=f"embed_node_task[{i}]",
                )
            else:
                self.nodes.append(n)
                self.node_map[n.doc_id] = n
                self.node_embeddings_list.append(n.embedding)

        embeddings = executor.results()
        for n, embedding in zip(nodes_to_embed, embeddings):
            n.embedding = embedding
            self.nodes.append(n)
            self.node_map[n.doc_id] = n
            self.node_embeddings_list.append(n.embedding)

    def get_node(self, node_id: str) -> Node:
        return self.node_map[node_id]

    def get_document(self, doc_id: str) -> Node:
        raise NotImplementedError

    def get_random_nodes(self, k=1) -> t.List[Node]:
        return choices(self.nodes, k=k)

    def get_similar(
        self, node: Node, threshold: float = 0.7, top_k: int = 3
    ) -> t.Union[t.List[Document], t.List[Node]]:
        items = []
        doc = node
        if doc.embedding is None:
            raise ValueError("Document has no embedding.")
        scores, doc_ids = get_top_k_embeddings(
            query_embedding=doc.embedding,
            embeddings=self.node_embeddings_list,
            similarity_fn=similarity,
            similarity_cutoff=threshold,
            # we need to return k+1 docs here as the top result is the input doc itself
            similarity_top_k=top_k + 1,
        )
        # remove the query doc itself from results
        scores, doc_ids = scores[1:], doc_ids[1:]
        items = [self.nodes[doc_id] for doc_id in doc_ids]
        return items

    def get_adjacent(
        self, node: Node, direction: Direction = Direction.NEXT
    ) -> t.Optional[Node]:
        # linear search for doc_id of doc in documents_list
        index = self.nodes.index(node)

        if direction == Direction.NEXT:
            if len(self.nodes) > index + 1:
                next_doc = self.nodes[index + 1]
                if next_doc.filename == node.filename:
                    return next_doc
                else:
                    return None
            else:
                return None
        if direction == Direction.PREV:
            if index > 0:
                prev_doc = self.nodes[index - 1]
                if prev_doc.filename == node.filename:
                    return prev_doc
                else:
                    return None
            else:
                return None



from __future__ import annotations

import json
import logging
import typing as t
from dataclasses import dataclass

logger = logging.getLogger(__name__)

if t.TYPE_CHECKING:
    from langchain_core.callbacks import Callbacks

    from ragas.llms.base import BaseRagasLLM


def load_as_json(text) -> t.Dict:
    """
    validate and return given text as json
    """

    try:
        return json.loads(text)
    except ValueError as e:
        logger.warn(f"Invalid json: {e}")
        return {}


# not migrating to Prompt format to avoid circular imports
JSON_PROMPT = """\
Rewrite the input into valid json

Input:
{{
    "name": "John Doe",
    "age": 30,
    "isStudent": false
    "address": {{
        "street": "123 Main St",
        "city": "Anytown",
        "state": "CA",
    }}
    "hobbies": ["reading", "swimming", "cycling"]
}}
Output:
{{
    "name": "John Doe",
    "age": 30,
    "isStudent": false,
    "address": {{
        "street": "123 Main St",
        "city": "Anytown",
        "state": "CA"
    }},
    "hobbies": ["reading", "swimming", "cycling"]
}}


Input:
{{
    "statement": "The Earth is also known as "Terra" "
}}
Output:
{{
    "statement": "The Earth is also known as 'Terra'"
}}

Input:
{input}

Output:
"""


@dataclass
class JsonLoader:
    max_retries: int = 2

    def safe_load(self, text: str, llm: BaseRagasLLM, callbacks: Callbacks = None):
        retry = 0
        while retry <= self.max_retries:
            try:
                start, end = self._find_outermost_json(text)
                return json.loads(text[start:end])
            except ValueError:
                from ragas.llms.prompt import PromptValue

                results = llm.generate_text(
                    PromptValue(prompt_str=JSON_PROMPT.format(input=text)),
                    n=1,
                    callbacks=callbacks,
                )
                text = results.generations[0][0].text
            retry += 1

        return {}

    async def asafe_load(
        self, text: str, llm: BaseRagasLLM, callbacks: Callbacks = None
    ):
        retry = 0
        while retry <= self.max_retries:
            try:
                start, end = self._find_outermost_json(text)
                return json.loads(text[start:end])
            except ValueError:
                from ragas.llms.prompt import PromptValue

                results = await llm.agenerate_text(
                    PromptValue(prompt_str=JSON_PROMPT.format(input=text)),
                    n=1,
                    callbacks=callbacks,
                )
                text = results.generations[0][0].text
            retry += 1

        return {}

    def _find_outermost_json(self, text):
        stack = []
        start_index = -1

        for i, char in enumerate(text):
            if char in "{[":
                if len(stack) == 0:
                    start_index = i
                stack.append(char)

            elif char in "}]":
                if len(stack) > 0:
                    last = stack.pop()
                    if (char == "}" and last != "{") or (char == "]" and last != "["):
                        # Mismatched closing brace/bracket, invalid JSON
                        break

                if len(stack) == 0 and start_index != -1:
                    # Found a valid outermost JSON
                    return (
                        start_index,
                        i + 1,
                    )  # Add 1 to include the closing brace/bracket in the range

        return -1, -1  # No valid JSON found


json_loader = JsonLoader()



from __future__ import annotations

import logging
import typing as t
from dataclasses import dataclass, field

import numpy as np

from ragas.llms.json_load import json_loader
from ragas.llms.prompt import Prompt
from ragas.metrics.base import EvaluationMode, MetricWithLLM

if t.TYPE_CHECKING:
    from langchain_core.callbacks import Callbacks

    from ragas.llms.prompt import PromptValue

logger = logging.getLogger(__name__)

LONG_FORM_ANSWER_PROMPT = Prompt(
    name="long_form_answer",
    instruction="Create one or more statements from each sentence in the given answer.",
    examples=[
        {
            "question": "Who was  Albert Einstein and what is he best known for?",
            "answer": "He was a German-born theoretical physicist, widely acknowledged to be one of the greatest and most influential physicists of all time. He was best known for developing the theory of relativity, he also made important contributions to the development of the theory of quantum mechanics.",
            "statements": {
                "statements": [
                    "Albert Einstein, a German-born theoretical physicist, is renowned for being one of the most influential physicists in history.",
                    "Albert Einstein was best known for his theory of relativity.",
                    "Einstein's contributions significantly advanced the field of quantum mechanics",
                    "Recognized globally, Einstein's work has profoundly impacted the scientific community",
                    "Einstein's groundbreaking theories continue to shape our understanding of physics today.",
                ]
            },
        },
        {
            "question": "Cadmium Chloride is slightly soluble in this chemical, it is also called what?",
            "answer": "alcohol",
            "statements": {
                "statements": ["Cadmium Chloride is slightly soluble in alcohol."]
            },
        },
        {
            "question": "Were Hitler and Benito Mussolini of the same nationality?",
            "answer": "Sorry, I can't provide answer to that question.",
            "statements": {"statements": []},
        },
    ],
    input_keys=["question", "answer"],
    output_key="statements",
    output_type="JSON",
)  # noqa: E501


NLI_STATEMENTS_MESSAGE = Prompt(
    name="nli_statements",
    instruction="Natural language inference. Use only 'Yes' (1), 'No' (0) and 'Null' (-1) as verdict.",
    examples=[
        {
            "context": """John is a student at XYZ University. He is pursuing a degree in Computer Science. He is enrolled in several courses this semester, including Data Structures, Algorithms, and Database Management. John is a diligent student and spends a significant amount of time studying and completing assignments. He often stays late in the library to work on his projects.""",
            "statements": """
            statement_1: John is majoring in Biology.
            statement_2: John is taking a course on Artificial Intelligence.
            statement_3: John is a dedicated student.
            statement_4: John has a part-time job.
            """,
            "answer": [
                {
                    "statement_1": "John is majoring in Biology.",
                    "reason": "John's major is explicitly mentioned as Computer Science. There is no information suggesting he is majoring in Biology.",
                    "verdict": "0",
                },
                {
                    "statement_2": "John is taking a course on Artificial Intelligence.",
                    "reason": "The context mentions the courses John is currently enrolled in, and Artificial Intelligence is not mentioned. Therefore, it cannot be deduced that John is taking a course on AI.",
                    "verdict": "0",
                },
                {
                    "statement_3": "John is a dedicated student.",
                    "reason": "The context states that he spends a significant amount of time studying and completing assignments. Additionally, it mentions that he often stays late in the library to work on his projects, which implies dedication.",
                    "verdict": "1",
                },
                {
                    "statement_4": "John has a part-time job.",
                    "reason": "There is no information given in the context about John having a part-time job.",
                    "verdict": "0",
                },
            ],
        },
        {
            "context": """Photosynthesis is a process used by plants, algae, and certain bacteria to convert light energy into chemical energy.""",
            "statements": """statement_1: Albert Einstein was a genius.""",
            "answer": {
                "statement_1": "Albert Einstein was a genius.",
                "reason": "The context and statement are unrelated",
                "verdict": "0",
            },
        },
        {
            "context": """Albert Einstein was a German-born theoretical physicist who is widely held to be one of the greatest and most influential scientists of all time.""",
            "statements": """statement_1: Nil""",
            "answer": {
                "statement_1": "Nil",
                "reason": "The statement is invalid",
                "verdict": "-1",
            },
        },
    ],
    input_keys=["context", "statements"],
    output_key="answer",
    output_type="JSON",
)  # noqa: E501


@dataclass
class Faithfulness(MetricWithLLM):
    name: str = "faithfulness"  # type: ignore
    evaluation_mode: EvaluationMode = EvaluationMode.qac  # type: ignore
    long_form_answer_prompt: Prompt = field(
        default_factory=lambda: LONG_FORM_ANSWER_PROMPT
    )
    nli_statements_message: Prompt = field(
        default_factory=lambda: NLI_STATEMENTS_MESSAGE
    )
    batch_size: int = 15

    def _create_answer_prompt(self, row: t.Dict) -> PromptValue:
        question, answer = row["question"], row["answer"]

        # extract statements from answer given the question
        prompt_value = self.long_form_answer_prompt.format(
            question=question, answer=answer
        )
        return prompt_value

    def _create_nli_prompt(self, row: t.Dict, statements: t.Any) -> PromptValue:
        assert self.llm is not None, "llm must be set to compute score"

        contexts = row["contexts"]
        # check if the statements are support in the contexts
        contexts_str: str = "\n".join(contexts)
        statements = statements if statements != [] else ["Nil"]
        statements_str: str = "\n".join(
            [f"statement_{i+1}: {st}" for i, st in enumerate(statements)]
        )
        prompt_value = self.nli_statements_message.format(
            context=contexts_str, statements=statements_str
        )
        return prompt_value

    def _compute_score(self, output: t.Any):
        # check the verdicts and compute the score
        verdict_score_map = {"1": 1, "0": 0, "null": np.nan}
        output = output if isinstance(output, list) else [output]
        faithful_statements = sum(
            verdict_score_map.get(
                statement_with_validation.get("verdict", "").lower(), np.nan
            )
            for statement_with_validation in output
        )
        num_statements = len(output)
        if num_statements:
            score = faithful_statements / num_statements
        else:
            score = np.nan

        return score

    async def _ascore(self: t.Self, row: t.Dict, callbacks: Callbacks) -> float:
        """
        returns the NLI score for each (q, c, a) pair
        """
        assert self.llm is not None, "LLM is not set"
        p = self._create_answer_prompt(row)
        answer_result = await self.llm.agenerate_text(p, callbacks=callbacks)

        statements = await json_loader.asafe_load(
            answer_result.generations[0][0].text, self.llm
        )
        p = self._create_nli_prompt(row, statements.get("statements", []))
        result = await self.llm.agenerate_text(p, callbacks=callbacks)

        json_output = await json_loader.asafe_load(
            result.generations[0][0].text, self.llm
        )
        return self._compute_score(json_output)

    def _score(self, row: t.Dict, callbacks: Callbacks) -> float:
        assert self.llm is not None, "LLM is not set"
        p = self._create_answer_prompt(row)
        answer_result = self.llm.generate_text(p, callbacks=callbacks)

        statements = json_loader.safe_load(
            answer_result.generations[0][0].text, self.llm
        )
        p = self._create_nli_prompt(row, statements.get("statements", []))
        result = self.llm.generate_text(p, callbacks=callbacks)

        json_output = json_loader.safe_load(result.generations[0][0].text, self.llm)
        return self._compute_score(json_output)

    def adapt(self, language: str, cache_dir: t.Optional[str] = None) -> None:
        assert self.llm is not None, "LLM is not set"

        logger.info(f"Adapting Faithfulness metric to {language}")
        self.long_form_answer_prompt = self.long_form_answer_prompt.adapt(
            language, self.llm, cache_dir
        )
        self.nli_statements_message = self.nli_statements_message.adapt(
            language, self.llm, cache_dir
        )

    def save(self, cache_dir: t.Optional[str] = None) -> None:
        self.long_form_answer_prompt.save(cache_dir)
        self.nli_statements_message.save(cache_dir)


faithfulness = Faithfulness()



from __future__ import annotations

import typing as t
from abc import ABC, abstractmethod
from dataclasses import dataclass

from langchain_community.chat_models import AzureChatOpenAI, ChatOpenAI, ChatVertexAI
from langchain_community.llms import AzureOpenAI, OpenAI, VertexAI
from langchain_core.language_models import BaseLanguageModel
from langchain_core.outputs import LLMResult

if t.TYPE_CHECKING:
    from langchain_core.callbacks import Callbacks

    from ragas.llms.prompt import PromptValue


MULTIPLE_COMPLETION_SUPPORTED = [
    OpenAI,
    ChatOpenAI,
    AzureOpenAI,
    AzureChatOpenAI,
    ChatVertexAI,
    VertexAI,
]


def is_multiple_completion_supported(llm: BaseLanguageModel) -> bool:
    """Return whether the given LLM supports n-completion."""
    for llm_type in MULTIPLE_COMPLETION_SUPPORTED:
        if isinstance(llm, llm_type):
            return True
    return False


@dataclass
class BaseRagasLLM(ABC):
    def get_temperature(self, n: int) -> float:
        """Return the temperature to use for completion based on n."""
        return 0.3 if n > 1 else 1e-8

    @abstractmethod
    def generate_text(
        self,
        prompt: PromptValue,
        n: int = 1,
        temperature: float = 1e-8,
        stop: t.Optional[t.List[str]] = None,
        callbacks: Callbacks = [],
    ) -> LLMResult:
        ...

    @abstractmethod
    async def agenerate_text(
        self,
        prompt: PromptValue,
        n: int = 1,
        temperature: float = 1e-8,
        stop: t.Optional[t.List[str]] = None,
        callbacks: Callbacks = [],
    ) -> LLMResult:
        ...


@dataclass
class LangchainLLMWrapper(BaseRagasLLM):
    """
    A simple base class for RagasLLMs that is based on Langchain's BaseLanguageModel
    interface. it implements 2 functions:
    - generate_text: for generating text from a given PromptValue
    - agenerate_text: for generating text from a given PromptValue asynchronously
    """

    langchain_llm: BaseLanguageModel

    def generate_text(
        self,
        prompt: PromptValue,
        n: int = 1,
        temperature: float = 1e-8,
        stop: t.Optional[t.List[str]] = None,
        callbacks: t.Optional[Callbacks] = None,
    ) -> LLMResult:
        temperature = self.get_temperature(n=n)
        if is_multiple_completion_supported(self.langchain_llm):
            return self.langchain_llm.generate_prompt(
                prompts=[prompt],
                n=n,
                temperature=temperature,
                stop=stop,
                callbacks=callbacks,
            )
        else:
            result = self.langchain_llm.generate_prompt(
                prompts=[prompt] * n,
                temperature=temperature,
                stop=stop,
                callbacks=callbacks,
            )
            # make LLMResult.generation appear as if it was n_completions
            # note that LLMResult.runs is still a list that represents each run
            generations = [[g[0] for g in result.generations]]
            result.generations = generations
            return result

    async def agenerate_text(
        self,
        prompt: PromptValue,
        n: int = 1,
        temperature: float = 1e-8,
        stop: t.Optional[t.List[str]] = None,
        callbacks: t.Optional[Callbacks] = None,
    ) -> LLMResult:
        temperature = self.get_temperature(n=n)
        if is_multiple_completion_supported(self.langchain_llm):
            return await self.langchain_llm.agenerate_prompt(
                prompts=[prompt],
                n=n,
                temperature=temperature,
                stop=stop,
                callbacks=callbacks,
            )
        else:
            result = await self.langchain_llm.agenerate_prompt(
                prompts=[prompt] * n,
                temperature=temperature,
                stop=stop,
                callbacks=callbacks,
            )
            # make LLMResult.generation appear as if it was n_completions
            # note that LLMResult.runs is still a list that represents each run
            generations = [[g[0] for g in result.generations]]
            result.generations = generations
            return result



from __future__ import annotations

import logging
import typing as t
from dataclasses import dataclass, field

import numpy as np

from ragas.llms.json_load import json_loader
from ragas.llms.prompt import Prompt
from ragas.metrics._answer_similarity import AnswerSimilarity
from ragas.metrics.base import EvaluationMode, MetricWithEmbeddings, MetricWithLLM

logger = logging.getLogger(__name__)

if t.TYPE_CHECKING:
    from langchain_core.callbacks import Callbacks

CORRECTNESS_PROMPT = Prompt(
    name="answer_correctness",
    instruction="""Extract following from given question and ground truth""",
    examples=[
        {
            "question": """What powers the sun and what is its primary function?""",
            "answer": """The sun is powered by nuclear fission, similar to nuclear reactors on Earth, and its primary function is to provide light to the solar system.""",
            "ground_truth": """The sun is actually powered by nuclear fusion, not fission. In its core, hydrogen atoms fuse to form helium, releasing a tremendous amount of energy. This energy is what lights up the sun and provides heat and light, essential for life on Earth. The sun's light also plays a critical role in Earth's climate system and helps to drive the weather and ocean currents.""",
            "Extracted statements": [
                {
                    "statements that are present in both the answer and the ground truth": [
                        "The sun's primary function is to provide light"
                    ],
                    "statements present in the answer but not found in the ground truth": [
                        "The sun is powered by nuclear fission",
                        "similar to nuclear reactors on Earth",
                    ],
                    "relevant statements found in the ground truth but omitted in the answer": [
                        "The sun is powered by nuclear fusion, not fission",
                        "In its core, hydrogen atoms fuse to form helium, releasing a tremendous amount of energy",
                        "This energy provides heat and light, essential for life on Earth",
                        "The sun's light plays a critical role in Earth's climate system",
                        "The sun helps to drive the weather and ocean currents",
                    ],
                }
            ],
        },
        {
            "question": """What is the boiling point of water?""",
            "answer": """The boiling point of water is 100 degrees Celsius at sea level.""",
            "ground_truth": """The boiling point of water is 100 degrees Celsius (212 degrees Fahrenheit) at sea level, but it can change with altitude.""",
            "Extracted statements": [
                {
                    "statements that are present in both the answer and the ground truth": [
                        "The boiling point of water is 100 degrees Celsius at sea level"
                    ],
                    "statements present in the answer but not found in the ground truth": [],
                    "relevant statements found in the ground truth but omitted in the answer": [
                        "The boiling point can change with altitude",
                        "The boiling point of water is 212 degrees Fahrenheit at sea level",
                    ],
                }
            ],
        },
    ],
    input_keys=["question", "answer", "ground_truth"],
    output_key="Extracted statements",
    output_type="json",
)


@dataclass
class AnswerCorrectness(MetricWithLLM, MetricWithEmbeddings):

    """
    Measures answer correctness compared to ground truth as a combination of
    factuality and semantic similarity.

    Attributes
    ----------
    name: string
        The name of the metrics
    batch_size: int
        batch size for evaluation
    weights:
        a list of two weights corresponding to factuality and semantic similarity
        Defaults [0.75, 0.25]
    answer_similarity:
        The AnswerSimilarity object
    """

    name: str = "answer_correctness"  # type: ignore[reportIncompatibleMethodOverride]
    evaluation_mode: EvaluationMode = EvaluationMode.qga  # type: ignore[reportIncompatibleMethodOverride]
    correctness_prompt: Prompt = field(default_factory=lambda: CORRECTNESS_PROMPT)
    batch_size: int = 15
    weights: list[float] = field(default_factory=lambda: [0.75, 0.25])
    answer_similarity: AnswerSimilarity | None = None

    def __post_init__(self: t.Self):
        if len(self.weights) != 2:
            raise ValueError(
                "Expects a list of two weights. First for factuality, second for semantic similarity"
            )
        if all([w == 0 for w in self.weights]):
            raise ValueError("At least one weight must be non-zero")
        if not all([w >= 0 for w in self.weights]):
            raise ValueError("Weights must be non-negative")

    def init_model(self):
        super().init_model()
        if self.answer_similarity is None and self.weights[1] != 0:
            self.answer_similarity = AnswerSimilarity(
                llm=self.llm, batch_size=self.batch_size, embeddings=self.embeddings
            )

    def _compute_statement_presence(self, prediction: t.Any) -> float:
        assert self.llm is not None, "LLM must be set"

        key_map = {
            "TP": "statements that are present in both the answer and the ground truth",
            "FP": "statements present in the answer but not found in the ground truth",
            "FN": "relevant statements found in the ground truth but omitted in the answer",  # noqa: E501
        }

        prediction = prediction if isinstance(prediction, list) else [prediction]
        if prediction:
            prediction = [
                item.get(key_map[k], np.nan)
                for item in prediction
                for k in key_map.keys()
            ]
            tp, fp, fn = [
                len(item) if isinstance(item, list) else np.nan for item in prediction
            ]
            if any([np.isnan(i) for i in [tp, fp, fn]]):
                score = np.nan
            else:
                score = tp / (tp + 0.5 * (fp + fn)) if tp > 0 else 0
        else:
            score = np.nan

        return score

    def _score(self, row: t.Dict, callbacks: Callbacks) -> float:
        assert self.llm is not None, "LLM must be set"
        q, a, g = row["question"], row["answer"], row["ground_truths"][0]
        p_value = self.correctness_prompt.format(question=q, ground_truth=g, answer=a)
        is_statement_present = self.llm.generate_text(p_value, callbacks=callbacks)

        prediction = json_loader.safe_load(
            is_statement_present.generations[0][0].text, self.llm
        )
        f1_score = self._compute_statement_presence(prediction)

        if self.weights[1] == 0:
            similarity_score = 0
        else:
            similarity_score = self.answer_similarity.score(row, callbacks=callbacks)  # type: ignore

        score = np.average(
            [f1_score, similarity_score],
            weights=self.weights,
        )

        return float(score)

    async def _ascore(self, row: t.Dict, callbacks: Callbacks) -> float:
        assert self.llm is not None, "LLM must be set"

        q, a, g = row["question"], row["answer"], row["ground_truths"][0]
        p_value = self.correctness_prompt.format(question=q, ground_truth=g, answer=a)
        is_statement_present = await self.llm.agenerate_text(
            p_value, callbacks=callbacks
        )

        prediction = await json_loader.asafe_load(
            is_statement_present.generations[0][0].text, self.llm
        )
        f1_score = self._compute_statement_presence(prediction)

        if self.weights[1] == 0:
            similarity_score = 0
        else:
            assert self.answer_similarity is not None, "AnswerSimilarity must be set"

            similarity_score = await self.answer_similarity.ascore(
                row, callbacks=callbacks
            )

        score = np.average(
            [f1_score, similarity_score],
            weights=self.weights,
        )

        return float(score)

    def adapt(self, language: str, cache_dir: t.Optional[str] = None) -> None:
        assert self.llm is not None, "llm must be set to compute score"

        logger.info(f"Adapting AnswerCorrectness metric to {language}")
        self.correctness_prompt = self.correctness_prompt.adapt(
            language, self.llm, cache_dir
        )

    def save(self, cache_dir: t.Optional[str] = None) -> None:
        self.correctness_prompt.save(cache_dir)


answer_correctness = AnswerCorrectness()



import os
import time

from llama_index import download_loader

from ragas.testset.evolutions import multi_context, reasoning, simple
from ragas.testset.generator import TestsetGenerator

generator = TestsetGenerator.with_openai()

distributions = {simple: 0.5, multi_context: 0.4, reasoning: 0.1}


def get_documents():
    SemanticScholarReader = download_loader("SemanticScholarReader")
    loader = SemanticScholarReader()
    # Narrow down the search space
    query_space = "large language models"
    # Increase the limit to obtain more documents
    documents = loader.load_data(query=query_space, limit=10)

    return documents


IGNORE_THREADS = True
IGNORE_ASYNCIO = False

if __name__ == "__main__":
    documents = get_documents()

    # asyncio
    if not IGNORE_ASYNCIO:
        os.environ["PYTHONASYNCIODEBUG"] = "1"
        print("Starting [Asyncio]")
        start = time.time()
        generator.generate_with_llamaindex_docs(
            documents=documents, test_size=100, distributions=distributions
        )
        print(f"Time taken: {time.time() - start:.2f}s")

    # Threads
    if not IGNORE_THREADS:
        print("Starting [Threads]")



from __future__ import annotations

import typing as t
from dataclasses import dataclass, field

import numpy as np
from datasets import Dataset, concatenate_datasets
from langchain_core.language_models import BaseLanguageModel

from ragas._analytics import EvaluationEvent, track
from ragas.callbacks import new_group
from ragas.embeddings.base import BaseRagasEmbeddings
from ragas.executor import Executor
from ragas.llms.base import BaseRagasLLM, LangchainLLMWrapper
from ragas.metrics.base import Metric, MetricWithEmbeddings, MetricWithLLM
from ragas.metrics.critique import AspectCritique

# from ragas.metrics.critique import AspectCritique
from ragas.validation import (
    remap_column_names,
    validate_column_dtypes,
    validate_evaluation_modes,
)

if t.TYPE_CHECKING:
    from langchain_core.callbacks import Callbacks


def evaluate(
    dataset: Dataset,
    metrics: list[Metric] | None = None,
    llm: t.Optional[BaseRagasLLM] = None,
    embeddings: t.Optional[BaseRagasEmbeddings] = None,
    callbacks: Callbacks = [],
    is_async: bool = False,
    max_workers: t.Optional[int] = None,
    raise_exceptions: bool = True,
    column_map: t.Dict[str, str] = {},
) -> Result:
    """
    Run the evaluation on the dataset with different metrics

    Parameters
    ----------
    dataset : Dataset[question: list[str], contexts: list[list[str]], answer: list[str], ground_truths: list[list[str]]]
        The dataset in the format of ragas which the metrics will use to score the RAG
        pipeline with
    metrics : list[Metric] , optional
        List of metrics to use for evaluation. If not provided then ragas will run the
        evaluation on the best set of metrics to give a complete view.
    llm: BaseRagasLLM, optional
        The language model to use for the metrics. If not provided then ragas will use
        the default language model. This can we overridden by the llm specified in
        the metric level with `metric.llm`.
    embeddings: BaseRagasEmbeddings, optional
        The embeddings to use for the metrics. If not provided then ragas will use
        the default embeddings. This can we overridden by the embeddings specified in
        the metric level with `metric.embeddings`.
    callbacks: Callbacks, optional
        Lifecycle Langchain Callbacks to run during evaluation. Check the
        [langchain documentation](https://python.langchain.com/docs/modules/callbacks/)
        for more information.
    is_async: bool, optional
        Whether to run the evaluation in async mode or not. If set to True then the
        evaluation is run by calling the `metric.ascore` method. In case the llm or
        embeddings does not support async then the evaluation can be run in sync mode
        with `is_async=False`. Default is False.
    max_workers: int, optional
        The number of workers to use for the evaluation. This is used by the
        `ThreadpoolExecutor` to run the evaluation in sync mode.
    raise_exceptions: bool, optional
        Whether to raise exceptions or not. If set to True then the evaluation will
        raise an exception if any of the metrics fail. If set to False then the
        evaluation will return `np.nan` for the row that failed. Default is True.
    column_map : dict[str, str], optional
        The column names of the dataset to use for evaluation. If the column names of
        the dataset are different from the default ones then you can provide the
        mapping as a dictionary here. Example: If the dataset column name is contexts_v1,
        column_map can be given as {"contexts":"contexts_v1"}

    Returns
    -------
    Result
        Result object containing the scores of each metric. You can use this do analysis
        later.

    Raises
    ------
    ValueError
        if validation fails because the columns required for the metrics are missing or
        if the columns are of the wrong format.

    Examples
    --------
    the basic usage is as follows:
    ```
    from ragas import evaluate

    >>> dataset
    Dataset({
        features: ['question', 'ground_truths', 'answer', 'contexts'],
        num_rows: 30
    })

    >>> result = evaluate(dataset)
    >>> print(result)
    {'context_precision': 0.817,
    'faithfulness': 0.892,
    'answer_relevancy': 0.874}
    ```
    """
    if dataset is None:
        raise ValueError("Provide dataset!")

    if metrics is None:
        from ragas.metrics import (
            answer_relevancy,
            context_precision,
            context_recall,
            faithfulness,
        )

        metrics = [answer_relevancy, context_precision, faithfulness, context_recall]
    # set the llm and embeddings
    if llm is None:
        from ragas.llms import llm_factory

        llm = llm_factory()
    elif isinstance(llm, BaseLanguageModel):
        llm = LangchainLLMWrapper(llm)
    if embeddings is None:
        from ragas.embeddings.base import embedding_factory

        embeddings = embedding_factory()

    # remap column names from the dataset
    dataset = remap_column_names(dataset, column_map)
    # validation
    validate_evaluation_modes(dataset, metrics)
    validate_column_dtypes(dataset)

    binary_metrics = []
    for metric in metrics:
        if isinstance(metric, AspectCritique):
            binary_metrics.append(metric.name)
        if isinstance(metric, MetricWithLLM):
            if metric.llm is None:
                metric.llm = llm
        if isinstance(metric, MetricWithEmbeddings):
            if metric.embeddings is None:
                metric.embeddings = embeddings

    # initialize all the models in the metrics
    [m.init_model() for m in metrics]

    executor = Executor(
        desc="Evaluating",
        keep_progress_bar=True,
        is_async=is_async,
        max_workers=max_workers,
        raise_exceptions=raise_exceptions,
    )
    # new evaluation chain
    row_run_managers = []
    evaluation_rm, evaluation_group_cm = new_group(
        name="ragas evaluation", inputs={}, callbacks=callbacks, is_async=is_async
    )
    for i, row in enumerate(dataset):
        row = t.cast(t.Dict[str, t.Any], row)
        row_rm, row_group_cm = new_group(
            name=f"row {i}",
            inputs=row,
            callbacks=evaluation_group_cm,
            is_async=is_async,
        )
        row_run_managers.append((row_rm, row_group_cm))

        if is_async:
            [
                executor.submit(
                    metric.ascore, row, row_group_cm, name=f"{metric.name}-{i}"
                )
                for metric in metrics
            ]
        else:
            [executor.submit(metric.score, row, row_group_cm) for metric in metrics]

    scores = []
    try:
        # get the results
        results = executor.results()
        # convert results to dataset_like
        for i, _ in enumerate(dataset):
            s = {}
            for j, m in enumerate(metrics):
                s[m.name] = results[len(metrics) * i + j]
            scores.append(s)
            # close the row chain
            row_rm, row_group_cm = row_run_managers[i]
            if not row_group_cm.ended:
                row_rm.on_chain_end(s)

    # run evaluation task
    except Exception as e:
        if not evaluation_group_cm.ended:
            evaluation_rm.on_chain_error(e)

        raise e
    else:
        result = Result(
            scores=Dataset.from_list(scores),
            dataset=dataset,
            binary_columns=binary_metrics,
        )
        if not evaluation_group_cm.ended:
            evaluation_rm.on_chain_end(result)

    # log the evaluation event
    metrics_names = [m.name for m in metrics]
    track(
        EvaluationEvent(
            event_type="evaluation",
            metrics=metrics_names,
            evaluation_mode="",
            num_rows=dataset.shape[0],
        )
    )
    return result


@dataclass
class Result(dict):
    scores: Dataset
    dataset: t.Optional[Dataset] = None
    binary_columns: t.List[str] = field(default_factory=list)

    def __post_init__(self):
        values = []
        for cn in self.scores[0].keys():
            value = np.nanmean(self.scores[cn])
            self[cn] = value
            if cn not in self.binary_columns:
                value = t.cast(float, value)
                values.append(value + 1e-10)

    def to_pandas(self, batch_size: int | None = None, batched: bool = False):
        if self.dataset is None:
            raise ValueError("dataset is not provided for the results class")
        assert self.scores.shape[0] == self.dataset.shape[0]
        result_ds = concatenate_datasets([self.dataset, self.scores], axis=1)

        return result_ds.to_pandas(batch_size=batch_size, batched=batched)

    def __repr__(self) -> str:
        scores = self.copy()
        score_strs = [f"'{k}': {v:0.4f}" for k, v in scores.items()]
        return "{" + ", ".join(score_strs) + "}"



from __future__ import annotations

import json
import logging
import os
import typing as t

from langchain_core.messages import BaseMessage, HumanMessage
from langchain_core.prompt_values import PromptValue as BasePromptValue
from langchain_core.pydantic_v1 import BaseModel, root_validator

from ragas.llms import BaseRagasLLM
from ragas.llms.json_load import json_loader
from ragas.utils import get_cache_dir

Example = t.Dict[str, t.Any]


class PromptValue(BasePromptValue):
    prompt_str: str

    def to_messages(self) -> t.List[BaseMessage]:
        """Return prompt as a list of Messages."""
        return [HumanMessage(content=self.to_string())]

    def to_string(self) -> str:
        return self.prompt_str


class Prompt(BaseModel):
    """
    Prompt is a class that represents a prompt for the ragas metrics.

    Prompt is a class that represents a prompt for the ragas metrics.

    Attributes:
        name (str): The name of the prompt.
        instruction (str): The instruction for the prompt.
        examples (List[Dict[str, Any]]): List of example inputs and outputs for the prompt.
        input_keys (List[str]): List of input variable names.
        output_key (str): The output variable name.
        output_type (str): The type of the output (default: "json").
        language (str): The language of the prompt (default: "en").
    """

    name: str
    instruction: str
    examples: t.List[Example] = []
    input_keys: t.List[str]
    output_key: str
    output_type: str = "json"
    language: str = "english"

    @root_validator
    def validate_prompt(cls, values: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
        """
        Validate the template string to ensure that it is in desired format.
        """
        if values.get("instruction") is None or values.get("instruction") == "":
            raise ValueError("instruction cannot be empty")
        if values.get("input_keys") is None or values.get("instruction") == []:
            raise ValueError("input_keys cannot be empty")
        if values.get("output_key") is None or values.get("output_key") == "":
            raise ValueError("output_key cannot be empty")

        if values.get("examples"):
            output_key = values["output_key"]
            for no, example in enumerate(values["examples"]):
                for inp_key in values["input_keys"]:
                    if inp_key not in example:
                        raise ValueError(
                            f"example {no+1} does not have the variable {inp_key} in the definition"
                        )
                if output_key not in example:
                    raise ValueError(
                        f"example {no+1} does not have the variable {output_key} in the definition"
                    )
                if values["output_type"].lower() == "json":
                    try:
                        if output_key in example:
                            if isinstance(example[output_key], str):
                                json.loads(example[output_key])
                    except ValueError as e:
                        raise ValueError(
                            f"{output_key} in example {no+1} is not in valid json format: {e}"
                        )

        return values

    def to_string(self) -> str:
        """
        Generate the prompt string from the variables.
        """
        prompt_str = self.instruction + "\n"

        if self.examples:
            # Format the examples to match the Langchain prompt template
            for example in self.examples:
                for key, value in example.items():
                    value = (
                        json.dumps(value, ensure_ascii=False).encode("utf8").decode()
                    )
                    value = (
                        value.replace("{", "{{").replace("}", "}}")
                        if self.output_type.lower() == "json"
                        else value
                    )
                    prompt_str += f"\n{key}: {value}"
                prompt_str += "\n"

        if self.input_keys:
            prompt_str += "".join(f"\n{key}: {{{key}}}" for key in self.input_keys)
        if self.output_key:
            prompt_str += f"\n{self.output_key}: \n"

        return prompt_str

    def get_example_str(self, example_no: int) -> str:
        """
        Get the example string from the example number.
        """
        if example_no >= len(self.examples):
            raise ValueError(f"example number {example_no} is out of range")
        example = self.examples[example_no]
        example_str = ""
        for key, value in example.items():
            value = json.dumps(value, ensure_ascii=False).encode("utf8").decode()
            value = (
                value.replace("{", "{{").replace("}", "}}")
                if self.output_type.lower() == "json"
                else value
            )
            example_str += f"\n{key}: {value}"
        return example_str

    def format(self, **kwargs: t.Any) -> PromptValue:
        """
        Format the Prompt object into a ChatPromptTemplate object to be used in metrics.
        """
        if set(self.input_keys) != set(kwargs.keys()):
            raise ValueError(
                f"Input variables {self.input_keys} do not match with the given parameters {list(kwargs.keys())}"
            )
        prompt = self.to_string()
        return PromptValue(prompt_str=prompt.format(**kwargs))

    def adapt(
        self, language: str, llm: BaseRagasLLM, cache_dir: t.Optional[str] = None
    ) -> Prompt:
        # TODO: Add callbacks
        cache_dir = cache_dir if cache_dir else get_cache_dir()
        if os.path.exists(os.path.join(cache_dir, language, f"{self.name}.json")):
            return self._load(language, self.name, cache_dir)

        prompts = []
        for example in self.examples:
            prompts.extend(
                [
                    str_translation.format(
                        translate_to=language, input=example.get(key)
                    )
                    for key in self.input_keys
                ]
            )
            prompts.append(
                json_translatation.format(
                    translate_to=language, input=example.get(self.output_key)
                )
                if self.output_type.lower() == "json"
                else str_translation.format(
                    translate_to=language, input=example.get(self.output_key)
                )
            )

        # NOTE: this is a slow loop, consider Executor to fasten this
        results = []
        for p in prompts:
            results.append(llm.generate_text(p).generations[0][0].text)
        per_example_items = len(self.input_keys) + 1
        grouped_results = [
            results[i : i + per_example_items]
            for i in range(0, len(results), per_example_items)
        ]
        assert len(grouped_results) == len(
            self.examples
        ), "examples and adapted examples must be of equal length"
        for i, example in enumerate(grouped_results):
            example_dict = {}
            example_dict.update(
                {k: v for k, v in zip(self.input_keys, example[: len(self.input_keys)])}
            )
            example_dict[self.output_key] = (
                json_loader.safe_load(example[-1], llm)
                if self.output_type.lower() == "json"
                else example[-1]
            )

            self.examples[i] = example_dict

        self.language = language

        # TODO:Validate the prompt after adaptation

        return self

    def save(self, cache_dir: t.Optional[str] = None) -> None:
        cache_dir = cache_dir if cache_dir else get_cache_dir()
        cache_dir = os.path.join(cache_dir, self.language)
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)

        cache_path = os.path.join(cache_dir, f"{self.name}.json")
        with open(cache_path, "w") as file:
            json.dump(self.dict(), file, indent=4)

    @classmethod
    def _load(cls, language: str, name: str, cache_dir: str) -> Prompt:
        logging.log(logging.INFO, f"Loading {name} from {cache_dir}")
        path = os.path.join(cache_dir, language, f"{name}.json")
        return cls(**json.load(open(path)))


str_translation = Prompt(
    name="str_translation",
    instruction="Language translation",
    examples=[
        {
            "translate_to": "hindi",
            "input": "Who was  Albert Einstein and what is he best known for?",
            "output": "अल्बर्ट आइंस्टीन कौन थे और वे किसके लिए सबसे ज्यादा प्रसिद्ध हैं?",
        },
    ],
    input_keys=["translate_to", "input"],
    output_key="output",
    output_type="str",
)

json_translatation = Prompt(
    name="json_translation",
    instruction="Translate values in given json to target language ",
    examples=[
        {
            "translate_to": "hindi",
            "input": """{"statements": [
            "Albert Einstein was born in Germany.",
            "Albert Einstein was best known for his theory of relativity."
        ]}""",
            "output": """{"statements": [
    "अल्बर्ट आइंस्टीन का जन्म जर्मनी में हुआ था।",
    "अल्बर्ट आइंस्टीन अपने सापेक्षता के सिद्धांत के लिए सबसे अधिक प्रसिद्ध थे।"
    ]}""",
        }
    ],
    input_keys=["translate_to", "input"],
    output_key="output",
    output_type="JSON",
)



from __future__ import annotations

import logging
import typing as t
from dataclasses import dataclass

import numpy as np

from ragas.embeddings.base import HuggingfaceEmbeddings
from ragas.metrics.base import EvaluationMode, MetricWithEmbeddings, MetricWithLLM

if t.TYPE_CHECKING:
    from langchain_core.callbacks.base import Callbacks


logger = logging.getLogger(__name__)


@dataclass
class AnswerSimilarity(MetricWithLLM, MetricWithEmbeddings):
    """
    Scores the semantic similarity of ground truth with generated answer.
    cross encoder score is used to quantify semantic similarity.
    SAS paper: https://arxiv.org/pdf/2108.06130.pdf

    Attributes
    ----------
    name : str
    batch_size : int
        Batch size.
    model_name:
        The model to be used for calculating semantic similarity
        Defaults open-ai-embeddings
        select cross-encoder model for best results
        https://huggingface.co/spaces/mteb/leaderboard
    threshold:
        The threshold if given used to map output to binary
        Default 0.5
    """

    name: str = "answer_similarity"  # type: ignore
    evaluation_mode: EvaluationMode = EvaluationMode.ga  # type: ignore
    batch_size: int = 15
    is_cross_encoder: bool = False
    threshold: t.Optional[float] = None

    def __post_init__(self: t.Self):
        # only for cross encoder
        if isinstance(self.embeddings, HuggingfaceEmbeddings):
            self.is_cross_encoder = True if self.embeddings.is_cross_encoder else False
            self.embeddings.encode_kwargs = {
                **self.embeddings.encode_kwargs,
                "batch_size": self.batch_size,
            }

    def init_model(self):
        super().init_model()

    def _score(self, row: t.Dict, callbacks: Callbacks) -> float:
        assert self.embeddings is not None, "embeddings must be set"

        ground_truths, answers = row["ground_truths"], row["answer"]
        ground_truths = [item[0] for item in ground_truths]

        if self.is_cross_encoder and isinstance(self.embeddings, HuggingfaceEmbeddings):
            raise NotImplementedError(
                "async score [ascore()] not implemented for HuggingFace embeddings"
            )
        else:
            embeddings_1 = np.array(self.embeddings.embed_documents(ground_truths))
            embeddings_2 = np.array(self.embeddings.embed_documents(answers))
            similarity = embeddings_1 @ embeddings_2.T
            if similarity.size == 1:
                # If similarity has only one value, directly use this value as scores
                scores = similarity.flatten()
            else:
                # If similarity contains multiple values, extract the diagonal as scores
                scores = np.diagonal(similarity)

        assert isinstance(scores, np.ndarray), "Expects ndarray"
        if self.threshold:
            scores = scores >= self.threshold  # type: ignore

        return scores.tolist()[0]

    async def _ascore(self: t.Self, row: t.Dict, callbacks: Callbacks = []) -> float:
        assert self.embeddings is not None, "embeddings must be set"

        ground_truths, answers = row["ground_truths"], row["answer"]
        ground_truths = [item[0] for item in ground_truths]

        if self.is_cross_encoder and isinstance(self.embeddings, HuggingfaceEmbeddings):
            raise NotImplementedError(
                "async score [ascore()] not implemented for HuggingFace embeddings"
            )
        else:
            embeddings_1 = np.array(
                await self.embeddings.aembed_documents(ground_truths)
            )
            embeddings_2 = np.array(await self.embeddings.aembed_documents(answers))
            similarity = embeddings_1 @ embeddings_2.T
            if similarity.size == 1:
                scores = similarity.flatten()
            else:
                scores = np.diagonal(similarity)

        assert isinstance(scores, np.ndarray), "Expects ndarray"
        if self.threshold:
            scores = scores >= self.threshold  # type: ignore

        return scores.tolist()[0]


answer_similarity = AnswerSimilarity()



from ragas import adapt
from ragas.metrics import context_recall


def test_adapt():
    adapt([context_recall], language="spanish")
    assert context_recall.context_recall_prompt.language == "spanish"



"""
Q - question
A - answer: generated_text from RAG pipeline
C - contexts: context used for generation
G - ground_truths: ground truth answer
"""
from __future__ import annotations

import typing as t
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum

from ragas.callbacks import new_group

if t.TYPE_CHECKING:
    from langchain_core.callbacks import Callbacks

    from ragas.embeddings import BaseRagasEmbeddings
    from ragas.llms import BaseRagasLLM


EvaluationMode = Enum("EvaluationMode", "qac qa qc gc ga qga qcg")


@dataclass
class Metric(ABC):
    batch_size: int

    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @property
    @abstractmethod
    def evaluation_mode(self) -> EvaluationMode:
        ...

    @abstractmethod
    def init_model(self):
        """
        This method will lazy initialize the model.
        """
        ...

    def adapt(self, language: str, cache_dir: t.Optional[str] = None) -> None:
        """
        Adapt the metric to a different language.
        """
        raise NotImplementedError(
            "adapt() is not implemented for {} metric".format(self.name)
        )

    def save(self, cache_dir: t.Optional[str] = None) -> None:
        """
        Save the metric to a path.
        """
        raise NotImplementedError(
            "adapt() is not implemented for {} metric".format(self.name)
        )

    def score(
        self: t.Self,
        row: t.Dict,
        callbacks: Callbacks = [],
    ) -> float:
        rm, group_cm = new_group(
            self.name, inputs=row, callbacks=callbacks, is_async=False
        )
        try:
            score = self._score(row=row, callbacks=group_cm)
        except Exception as e:
            if not group_cm.ended:
                rm.on_chain_error(e)
            raise e
        else:
            if not group_cm.ended:
                rm.on_chain_end({"output": score})
        return score

    @abstractmethod
    def _score(self, row: t.Dict, callbacks: Callbacks) -> float:
        ...

    async def ascore(self: t.Self, row: t.Dict, callbacks: Callbacks = []) -> float:
        rm, group_cm = new_group(
            self.name, inputs=row, callbacks=callbacks, is_async=True
        )
        try:
            score = await self._ascore(row=row, callbacks=group_cm)
        except Exception as e:
            if not group_cm.ended:
                rm.on_chain_error(e)
            raise e
        else:
            if not group_cm.ended:
                rm.on_chain_end({"output": score})
        return score

    @abstractmethod
    async def _ascore(self, row: t.Dict, callbacks: Callbacks) -> float:
        ...


@dataclass
class MetricWithLLM(Metric):
    llm: t.Optional[BaseRagasLLM] = None

    def init_model(self):
        """
        Init any models in the metric, this is invoked before evaluate()
        to load all the models
        Also check if the api key is valid for OpenAI and AzureOpenAI
        """
        if self.llm is None:
            raise ValueError(
                f"Metric '{self.name}' has no valid LLM provided (self.llm is None). Please initantiate a the metric with an LLM to run."  # noqa
            )


@dataclass
class MetricWithEmbeddings(Metric):
    embeddings: t.Optional[BaseRagasEmbeddings] = None

    def init_model(self):
        """
        Init any models in the metric, this is invoked before evaluate()
        to load all the models
        Also check if the api key is valid for OpenAI and AzureOpenAI
        """
        if self.embeddings is None:
            raise ValueError(
                f"Metric '{self.name}' has no valid embeddings provided (self.embeddings is None). Please initantiate a the metric with an embeddings to run."  # noqa
            )



from __future__ import annotations

from datasets import Dataset, Sequence

from ragas.metrics._context_precision import ContextPrecision
from ragas.metrics.base import EvaluationMode, Metric


def remap_column_names(dataset: Dataset, column_map: dict[str, str]) -> Dataset:
    """
    Remap the column names in case dataset uses different column names
    """

    inverse_column_map = {v: k for k, v in column_map.items()}
    return dataset.rename_columns(inverse_column_map)


def validate_column_dtypes(ds: Dataset):
    for column_names in ["question", "answer"]:
        if column_names in ds.features:
            if ds.features[column_names].dtype != "string":
                raise ValueError(
                    f'Dataset feature "{column_names}" should be of type string'
                )

    for column_names in ["contexts", "ground_truths"]:
        if column_names in ds.features:
            if not (
                isinstance(ds.features[column_names], Sequence)
                and ds.features[column_names].feature.dtype == "string"
            ):
                raise ValueError(
                    f'Dataset feature "{column_names}" should be of type'
                    f" Sequence[string], got {type(ds.features[column_names])}"
                )


EVALMODE_TO_COLUMNS = {
    EvaluationMode.qac: ["question", "answer", "contexts"],
    EvaluationMode.qa: ["question", "answer"],
    EvaluationMode.qc: ["question", "contexts"],
    EvaluationMode.gc: ["ground_truths", "contexts"],
    EvaluationMode.ga: ["ground_truths", "answer"],
    EvaluationMode.qga: ["question", "ground_truths", "answer"],
    EvaluationMode.qcg: ["question", "contexts", "ground_truths"],
}


def validate_evaluation_modes(ds: Dataset, metrics: list[Metric]):
    """
    validates the dataset and returns the evaluation type

    possible evaluation types
    1. (q,a,c)
    2. (q,a)
    3. (q,c)
    4. (g,a)
    """

    for m in metrics:
        required_columns = set(EVALMODE_TO_COLUMNS[m.evaluation_mode])
        available_columns = set(ds.features.keys())
        if not required_columns.issubset(available_columns):
            extra_msg = ""
            if (
                isinstance(m, ContextPrecision)
                and "ground_truths" not in available_columns
            ):
                extra_msg = "Looks like you're trying to use 'context_precision' without ground_truths. Please use consider using  `context_utilization' instead."

            raise ValueError(
                f"The metric [{m.name}] that that is used requires the following "
                f"additional columns {list(required_columns - available_columns)} "
                f"to be present in the dataset. {extra_msg}"
            )



from __future__ import annotations

import asyncio
import logging
import typing as t
from abc import ABC
from dataclasses import dataclass

from ragas.llms.json_load import load_as_json
from ragas.testset.prompts import (
    context_scoring_prompt,
    evolution_elimination_prompt,
    filter_question_prompt,
)

if t.TYPE_CHECKING:
    from ragas.llms.base import BaseRagasLLM
    from ragas.testset.docstore import Node


logger = logging.getLogger(__name__)


@dataclass
class Filter(ABC):
    ...


@dataclass
class NodeFilter(Filter):
    llm: BaseRagasLLM
    threshold: float = 7.5

    def filter(self, node: Node) -> t.Dict:
        return asyncio.get_event_loop().run_until_complete(self.afilter(node))

    async def afilter(self, node: Node) -> t.Dict:
        prompt = context_scoring_prompt.format(context=node.page_content)
        results = await self.llm.agenerate_text(prompt=prompt)
        output = results.generations[0][0].text.strip()
        score = load_as_json(output)
        score.update({"score": score.get("score", 0) >= self.threshold})
        return score


@dataclass
class QuestionFilter(Filter):
    llm: BaseRagasLLM

    def filter(self, question: str) -> bool:
        return asyncio.get_event_loop().run_until_complete(self.afilter(question))

    async def afilter(self, question: str) -> bool:
        prompt = filter_question_prompt.format(question=question)
        results = await self.llm.agenerate_text(prompt=prompt)
        results = results.generations[0][0].text.strip()
        json_results = load_as_json(results)
        logger.debug("filtered question: %s", json_results)
        return json_results.get("verdict") != "No"


@dataclass
class EvolutionFilter(Filter):
    llm: BaseRagasLLM

    def filter(self, simple_question: str, compressed_question: str) -> bool:
        return asyncio.get_event_loop().run_until_complete(
            self.afilter(simple_question, compressed_question)
        )

    async def afilter(self, simple_question: str, compressed_question: str) -> bool:
        prompt = evolution_elimination_prompt.format(
            question1=simple_question, question2=compressed_question
        )
        results = await self.llm.agenerate_text(prompt=prompt)
        results = results.generations[0][0].text.strip()
        json_results = load_as_json(results)
        logger.debug("filtered question: %s", json_results)
        return json_results.get("verdict") != "No"



from datasets import load_dataset

from ragas import evaluate
from ragas.metrics import answer_relevancy, context_precision, faithfulness
from ragas.metrics.critique import harmfulness


def test_evaluate_e2e():
    ds = load_dataset("explodinggradients/fiqa", "ragas_eval")["baseline"]
    result = evaluate(
        ds.select(range(3)),
        metrics=[answer_relevancy, context_precision, faithfulness, harmfulness],
    )
    assert result is not None



from __future__ import annotations

import logging
import typing as t
from dataclasses import dataclass

import pandas as pd
from langchain_openai.chat_models import ChatOpenAI
from langchain_openai.embeddings import OpenAIEmbeddings

from ragas._analytics import TesetGenerationEvent, track
from ragas.embeddings import BaseRagasEmbeddings
from ragas.executor import Executor
from ragas.llms import BaseRagasLLM, LangchainLLMWrapper
from ragas.testset.docstore import Document, DocumentStore, InMemoryDocumentStore
from ragas.testset.evolutions import (
    ComplexEvolution,
    CurrentNodes,
    DataRow,
    multi_context,
    reasoning,
    simple,
)
from ragas.testset.filters import EvolutionFilter, NodeFilter, QuestionFilter

if t.TYPE_CHECKING:
    from llama_index.readers.schema import Document as LlamaindexDocument
    from langchain_core.documents import Document as LCDocument

logger = logging.getLogger(__name__)

Distributions = t.Dict[t.Any, float]
DEFAULT_DISTRIBUTION = {simple: 0.5, reasoning: 0.25, multi_context: 0.25}


@dataclass
class TestDataset:
    """
    TestDataset class
    """

    test_data: t.List[DataRow]

    def to_pandas(self) -> pd.DataFrame:
        data_samples = []
        for data in self.test_data:
            data_dict = dict(data)
            data_dict["episode_done"] = True
            data_samples.append(data_dict)

        return pd.DataFrame.from_records(data_samples)


@dataclass
class TestsetGenerator:
    generator_llm: BaseRagasLLM
    critic_llm: BaseRagasLLM
    embeddings: BaseRagasEmbeddings
    docstore: DocumentStore

    @classmethod
    def with_openai(
        cls,
        generator_llm: str = "gpt-3.5-turbo",
        critic_llm: str = "gpt-4",
        embeddings: str = "text-embedding-ada-002",
        docstore: t.Optional[DocumentStore] = None,
        chunk_size: int = 512,
    ) -> "TestsetGenerator":
        generator_llm_model = LangchainLLMWrapper(ChatOpenAI(model=generator_llm))
        critic_llm_model = LangchainLLMWrapper(ChatOpenAI(model=critic_llm))
        embeddings_model = OpenAIEmbeddings(model=embeddings)
        if docstore is None:
            from langchain.text_splitter import TokenTextSplitter

            splitter = TokenTextSplitter(chunk_size=chunk_size, chunk_overlap=0)
            docstore = InMemoryDocumentStore(
                splitter=splitter, embeddings=embeddings_model
            )
            return cls(
                generator_llm=generator_llm_model,
                critic_llm=critic_llm_model,
                embeddings=embeddings_model,
                docstore=docstore,
            )
        else:
            return cls(
                generator_llm=generator_llm_model,
                critic_llm=critic_llm_model,
                embeddings=embeddings_model,
                docstore=docstore,
            )

    # if you add any arguments to this function, make sure to add them to
    # generate_with_langchain_docs as well
    def generate_with_llamaindex_docs(
        self,
        documents: t.Sequence[LlamaindexDocument],
        test_size: int,
        distributions: Distributions = {},
        with_debugging_logs=False,
    ):
        # chunk documents and add to docstore
        self.docstore.add_documents(
            [Document.from_llamaindex_document(doc) for doc in documents]
        )

        return self.generate(
            test_size=test_size,
            distributions=distributions,
            with_debugging_logs=with_debugging_logs,
        )

    # if you add any arguments to this function, make sure to add them to
    # generate_with_langchain_docs as well
    def generate_with_langchain_docs(
        self,
        documents: t.Sequence[LCDocument],
        test_size: int,
        distributions: Distributions = {},
        with_debugging_logs=False,
    ):
        # chunk documents and add to docstore
        self.docstore.add_documents(
            [Document.from_langchain_document(doc) for doc in documents]
        )

        return self.generate(
            test_size=test_size,
            distributions=distributions,
            with_debugging_logs=with_debugging_logs,
        )

    def generate(
        self,
        test_size: int,
        distributions: Distributions = DEFAULT_DISTRIBUTION,
        with_debugging_logs=False,
    ):
        # init filters and evolutions
        for evolution in distributions:
            if evolution.generator_llm is None:
                evolution.generator_llm = self.generator_llm
            if evolution.docstore is None:
                evolution.docstore = self.docstore

            if evolution.question_filter is None:
                evolution.question_filter = QuestionFilter(llm=self.critic_llm)
            if evolution.node_filter is None:
                evolution.node_filter = NodeFilter(llm=self.critic_llm)

            if isinstance(evolution, ComplexEvolution):
                evolution.init_evolution()
                if evolution.evolution_filter is None:
                    evolution.evolution_filter = EvolutionFilter(llm=self.critic_llm)
        if with_debugging_logs:
            from ragas.utils import patch_logger

            patch_logger("ragas.testset.evolutions", logging.DEBUG)

        exec = Executor(
            desc="Generating",
            keep_progress_bar=True,
            raise_exceptions=True,
            is_async=True,
        )

        current_nodes = [
            CurrentNodes(root_node=n, nodes=[n])
            for n in self.docstore.get_random_nodes(k=test_size)
        ]
        for evolution, probability in distributions.items():
            for i in range(round(probability * test_size)):
                exec.submit(
                    evolution.aevolve,
                    current_nodes[i],
                    name=f"{evolution.__class__.__name__}-{i}",
                )

        try:
            test_data_rows = exec.results()
        except ValueError as e:
            raise e
        test_dataset = TestDataset(test_data=test_data_rows)
        track(
            TesetGenerationEvent(
                event_type="testset_generation",
                evolutions={
                    k.__class__.__name__.lower(): v for k, v in distributions.items()
                },
                num_rows=len(test_dataset.test_data),
            )
        )

        return test_dataset



from __future__ import annotations

import typing as t
from dataclasses import field
from typing import List

import numpy as np
from langchain_core.embeddings import Embeddings as BaseRagasEmbeddings
from langchain_openai.embeddings import OpenAIEmbeddings
from pydantic.dataclasses import dataclass

DEFAULT_MODEL_NAME = "BAAI/bge-small-en-v1.5"


@dataclass
class HuggingfaceEmbeddings(BaseRagasEmbeddings):
    model_name: str = DEFAULT_MODEL_NAME
    """Model name to use."""
    cache_folder: t.Optional[str] = None
    """Path to store models. 
    Can be also set by SENTENCE_TRANSFORMERS_HOME environment variable."""
    model_kwargs: t.Dict[str, t.Any] = field(default_factory=dict)
    """Keyword arguments to pass to the model."""
    encode_kwargs: t.Dict[str, t.Any] = field(default_factory=dict)

    def __post_init__(self):
        try:
            import sentence_transformers
            from transformers import AutoConfig
            from transformers.models.auto.modeling_auto import (
                MODEL_FOR_SEQUENCE_CLASSIFICATION_MAPPING_NAMES,
            )
        except ImportError as exc:
            raise ImportError(
                "Could not import sentence_transformers python package. "
                "Please install it with `pip install sentence-transformers`."
            ) from exc
        config = AutoConfig.from_pretrained(self.model_name)
        self.is_cross_encoder = bool(
            np.intersect1d(
                list(MODEL_FOR_SEQUENCE_CLASSIFICATION_MAPPING_NAMES.values()),
                config.architectures,
            )
        )

        if self.is_cross_encoder:
            self.model = sentence_transformers.CrossEncoder(
                self.model_name, **self.model_kwargs
            )
        else:
            self.model = sentence_transformers.SentenceTransformer(
                self.model_name, cache_folder=self.cache_folder, **self.model_kwargs
            )

        # ensure outputs are tensors
        if "convert_to_tensor" not in self.encode_kwargs:
            self.encode_kwargs["convert_to_tensor"] = True

    def embed_query(self, text: str) -> List[float]:
        return self.embed_documents([text])[0]

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        from sentence_transformers.SentenceTransformer import SentenceTransformer
        from torch import Tensor

        assert isinstance(
            self.model, SentenceTransformer
        ), "Model is not of the type Bi-encoder"
        embeddings = self.model.encode(
            texts, normalize_embeddings=True, **self.encode_kwargs
        )

        assert isinstance(embeddings, Tensor)
        return embeddings.tolist()

    def predict(self, texts: List[List[str]]) -> List[List[float]]:
        from sentence_transformers.cross_encoder import CrossEncoder
        from torch import Tensor

        assert isinstance(
            self.model, CrossEncoder
        ), "Model is not of the type CrossEncoder"

        predictions = self.model.predict(texts, **self.encode_kwargs)

        assert isinstance(predictions, Tensor)
        return predictions.tolist()


def embedding_factory() -> BaseRagasEmbeddings:
    openai_embeddings = OpenAIEmbeddings()
    return openai_embeddings
